import logging
from datetime import datetime

logger = logging.getLogger(__name__)


def validate_date(d):
    try:
        return datetime.strptime(d, '%Y-%m-%d')
    except ValueError:
        return False



import logging
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path

from tqdm import tqdm

from deemon.cmd import search
from deemon.cmd.refresh import Refresh
from deemon.core.api import PlatformAPI
from deemon.core.config import Config as config
from deemon.core.db import Database
from deemon.utils import dataprocessor, ui

logger = logging.getLogger(__name__)


class Monitor:

    def __init__(self):
        self.bitrate = None
        self.alerts = False
        self.record_type = None
        self.download_path = None
        self.remove = False
        self.refresh = True
        self.is_search = False
        self.duplicates = 0
        self.time_machine = None
        self.dl = None
        self.db = Database()
        self.api = PlatformAPI("deezer-gw")
        self.artist_not_found = []

    def set_config(self, bitrate: str, alerts: bool, record_type: str, download_path: Path):
        self.bitrate = bitrate
        self.alerts = alerts
        self.record_type = record_type
        self.download_path = download_path
        self.debugger("SetConfig", {'bitrate': bitrate, 'alerts': alerts, 'type': record_type, 'path': download_path})

    def set_options(self, remove, dl_all, search):
        self.remove = True if remove else False
        self.dl = True if dl_all else False
        self.is_search = True if search else False
        self.debugger("SetOptions", {'remove': remove, 'dl': dl_all, 'search': search})

    def debugger(self, message: str, payload=None):
        if config.debug_mode():
            if not payload:
                payload = ""
            logger.debug(f"DEBUG_MODE: {message} {str(payload)}")

    def get_best_result(self, api_result) -> list:
        name = api_result['query']

        if self.is_search:
            logger.debug("Waiting for user input...")
            prompt = self.prompt_search(name, api_result['results'])
            if prompt:
                logger.debug(f"User selected {prompt}")
                return [prompt]

        matches = [r for r in api_result['results'] if r['name'].lower() == name.lower()]
        self.debugger("Matches", matches)

        if len(matches) == 1:
            return [matches[0]]
        elif len(matches) > 1:
            logger.debug(f"Multiple matches were found for artist \"{api_result['query']}\"")
            if config.prompt_duplicates():
                logger.debug("Waiting for user input...")
                prompt = self.prompt_search(name, matches)
                if prompt:
                    logger.debug(f"User selected {prompt}")
                    return [prompt]
                else:
                    logger.info(f"No selection made, skipping {name}...")
                    return []
            else:
                self.duplicates += 1
                return [matches[0]]
        elif not len(matches):
            logger.debug(f"No matches were found for artist \"{api_result['query']}\"")
            if config.prompt_no_matches():
                logger.debug("Waiting for user input...")
                prompt = self.prompt_search(name, api_result['results'])
                if prompt:
                    logger.debug(f"User selected {prompt}")
                    return [prompt]
                else:
                    logger.info(f"No selection made, skipping {name}...")
                    return []
            else:
                self.artist_not_found.append(name)
                return []

    def prompt_search(self, value, api_result):
        menu = search.Search()
        ask_user = menu.artist_menu(value, api_result, True)
        if ask_user:
            return {'id': ask_user['id'], 'name': ask_user['name']}
        return logger.debug("No artist selected, skipping...")

    # @performance.timeit
    def build_artist_query(self, api_result: list):
        existing = self.db.get_all_monitored_artist_ids()
        artists_to_add = []
        pbar = tqdm(api_result, total=len(api_result), desc="Setting up artists for monitoring...", ascii=" #",
                    bar_format=ui.TQDM_FORMAT)
        for artist in pbar:
            if artist['id'] in existing:
                logger.info(f"   - Already monitoring {artist['name']}, skipping...")
            else:
                artist.update({'bitrate': self.bitrate, 'alerts': self.alerts, 'record_type': self.record_type,
                               'download_path': self.download_path, 'profile_id': config.profile_id(),
                               'trans_id': config.transaction_id()})
                artists_to_add.append(artist)
        if len(artists_to_add):
            logger.debug("New artists have been monitored. Saving changes to the database...")
            self.db.new_transaction()
            self.db.fast_monitor(artists_to_add)
            self.db.commit()
            return True

    def build_playlist_query(self, api_result: list):
        existing = self.db.get_all_monitored_playlist_ids() or []
        playlists_to_add = []
        pbar = tqdm(api_result, total=len(api_result), desc="Setting up playlists for monitoring...", ascii=" #",
                    bar_format=ui.TQDM_FORMAT)
        for i, playlist in enumerate(pbar):
            if playlist['id'] in existing:
                logger.info(f"   Already monitoring {playlist['title']}, skipping...")
            else:
                playlist.update({'bitrate': self.bitrate, 'alerts': self.alerts, 'download_path': self.download_path,
                                 'profile_id': config.profile_id(), 'trans_id': config.transaction_id()})
                playlists_to_add.append(playlist)
        if len(playlists_to_add):
            logger.debug("New playlists have been monitored. Saving changes to the database...")
            self.db.new_transaction()
            self.db.fast_monitor_playlist(playlists_to_add)
            self.db.commit()
            return True

    def call_refresh(self):
        refresh = Refresh(self.time_machine, ignore_filters=self.dl)
        refresh.run()

    # @performance.timeit
    def artists(self, names: list) -> None:
        """
        Return list of dictionaries containing each artist
        """
        if self.remove:
            return self.purge_artists(names=names)
        self.debugger("SpawningThreads", self.api.max_threads)
        with ThreadPoolExecutor(max_workers=self.api.max_threads) as ex:
            api_result = list(
                tqdm(ex.map(self.api.search_artist, names), total=len(names),
                     desc=f"Fetching artist data for {len(names):,} artist(s), please wait...",
                     ascii=" #", bar_format=ui.TQDM_FORMAT))

        select_artist = tqdm(api_result, total=len(api_result), desc="Examining results for best match...", ascii=" #",
                             bar_format=ui.TQDM_FORMAT)

        to_monitor = []
        for artist in select_artist:
            best_result = self.get_best_result(artist)
            if best_result:
                to_monitor.append(best_result)

        to_process = [item for elem in to_monitor for item in elem if len(item)]
        if self.build_artist_query(to_process):
            self.call_refresh()
        else:
            print("")
            logger.info("No new artists have been added, skipping refresh.")

    # @performance.timeit
    def artist_ids(self, ids: list):
        ids = [int(x) for x in ids]
        if self.remove:
            return self.purge_artists(ids=ids)
        self.debugger("SpawningThreads", self.api.max_threads)
        with ThreadPoolExecutor(max_workers=self.api.max_threads) as ex:
            api_result = list(
                tqdm(ex.map(self.api.get_artist_by_id, ids), total=len(ids),
                     desc=f"Fetching artist data for {len(ids):,} artist(s), please wait...",
                     ascii=" #", bar_format=ui.TQDM_FORMAT))

        if self.build_artist_query(api_result):
            self.call_refresh()
        else:
            print("")
            logger.info("No new artists have been added, skipping refresh.")

    # @performance.timeit
    def importer(self, import_path: str):
        if Path(import_path).is_file():
            imported_file = dataprocessor.read_file_as_csv(import_path)
            artist_list = dataprocessor.process_input_file(imported_file)
            if isinstance(artist_list[0], int):
                self.artist_ids(artist_list)
            else:
                self.artists(artist_list)
        elif Path(import_path).is_dir():
            import_list = [x.relative_to(import_path).name for x in sorted(Path(import_path).iterdir()) if x.is_dir()]
            if import_list:
                self.artists(import_list)
        else:
            logger.error(f"File or directory not found: {import_path}")
            return

    # @performance.timeit
    def playlists(self, playlists: list):
        if self.remove:
            return self.purge_playlists(ids=playlists)
        ids = [int(x) for x in playlists]
        self.debugger("SpawningThreads", self.api.max_threads)
        with ThreadPoolExecutor(max_workers=self.api.max_threads) as ex:
            api_result = list(
                tqdm(ex.map(self.api.get_playlist, ids), total=len(ids),
                     desc=f"Fetching playlist data for {len(ids):,} playlist(s), please wait...",
                     ascii=" #", bar_format=ui.TQDM_FORMAT))

        if self.build_playlist_query(api_result):
            self.call_refresh()
        else:
            print("")
            logger.info("No new playlists have been added, skipping refresh.")

    # @performance.timeit
    def purge_artists(self, names: list = None, ids: list = None):
        if names:
            for n in names:
                monitored = self.db.get_monitored_artist_by_name(n)
                if monitored:
                    self.db.remove_monitored_artist(monitored['artist_id'])
                    logger.info(f"No longer monitoring {monitored['artist_name']}")
                else:
                    logger.info(f"{n} is not being monitored yet")
        if ids:
            for i in ids:
                monitored = self.db.get_monitored_artist_by_id(i)
                if monitored:
                    self.db.remove_monitored_artist(monitored['artist_id'])
                    logger.info(f"No longer monitoring {monitored['artist_name']}")
                else:
                    logger.info(f"{i} is not being monitored yet")

    def purge_playlists(self, titles: list = None, ids: list = None):
        if ids:
            for i in ids:
                monitored = self.db.get_monitored_playlist_by_id(i)
                if monitored:
                    self.db.remove_monitored_playlists(monitored['id'])
                    logger.info(f"No longer monitoring {monitored['title']}")
                else:
                    logger.info(f"{i} is not being monitored yet")



import logging

from deemon.core.db import Database
from deemon.utils import dates

logger = logging.getLogger(__name__)
db = Database()


def view_transactions():
    transactions = db.get_transactions()
    if not transactions:
        return logger.info("No transactions are available to be rolled back.")

    for i, transaction in enumerate(transactions, start=1):
        release_id = []
        artist_names = []
        playlist_titles = []

        for k, v in transaction.items():
            if (k == "releases" or k == "playlist_tracks") and transaction[k]:
                for item in transaction[k]:
                    for key, val in item.items():
                        if key == "album_id" or key == "track_id":
                            if item[key] not in release_id:
                                release_id.append(item[key])
            if k == "monitor" and transaction[k]:
                if transaction[k] not in artist_names:
                    artist_names = [x['artist_name'] for x in transaction[k]]
            if k == "playlists" and transaction[k]:
                if transaction[k] not in playlist_titles:
                    playlist_titles.append(transaction[k][0]['title'])

        release_id_len = len(release_id)

        if release_id_len == 1:
            release_text = f" and {release_id_len} release"
        elif release_id_len > 1:
            release_text = f" and {release_id_len} releases"
        else:
            release_text = ""

        if len(artist_names) > 1:
            if len(playlist_titles) > 1:
                playlist_text = f", {len(playlist_titles)} playlists"
            elif len(playlist_titles) == 1:
                playlist_text = f", {len(playlist_titles)} playlist"
            else:
                playlist_text = ""
            output_text = f"Added {artist_names[0]} + {len(artist_names) - 1} artist(s){playlist_text}{release_text}"
        elif len(artist_names) == 1:
            if len(playlist_titles) > 1:
                playlist_text = f", {len(playlist_titles)} playlists"
            elif len(playlist_titles) == 1:
                playlist_text = f", {len(playlist_titles)} playlist"
            else:
                playlist_text = ""
            output_text = f"Added {artist_names[0]}{playlist_text}{release_text}"
        else:
            if len(playlist_titles) > 1:
                output_text = f"Added {playlist_titles[0]} + {len(playlist_titles) - 1}{release_text}"
            elif len(playlist_titles) == 1:
                output_text = f"Added {playlist_titles[0]}{release_text}"
            else:
                output_text = f"Added {release_text[4:]}"

        print(f"{i}. {dates.get_friendly_date(transaction['timestamp'])} - {output_text}")

    rollback = None
    while rollback not in range(len(transactions)):
        rollback = input("\nSelect specific refresh to rollback (or press Enter to exit): ")
        if rollback == "":
            return
        try:
            rollback = int(rollback) - 1
        except ValueError:
            logger.error("Invalid input")

    rollback = transactions[rollback]['id']
    logger.debug(f"Rolling back transaction {rollback}")
    db.rollback_refresh(rollback)


def rollback_last(i: int):
    db.rollback_last_refresh(i)
    logger.info(f"Rolled back the last {i} transaction(s).")



import os


TQDM_FORMAT = ":: {desc} {percentage:3.0f}%"


def get_progress_bar_size() -> int:
    try:
        screen_size = int(os.get_terminal_size().columns)
    except OSError:
        screen_size = 80

    dynamic_size = int(screen_size / 4)
    if dynamic_size > 30:
        return 30
    elif dynamic_size < 16:
        return 16
    else:
        return dynamic_size


def set_progress_bar_text(msg: str, max_length: int) -> str:
    max_cols = get_progress_bar_size()
    max_length += 11

    if max_length < max_cols:
        max_cols = max_length

    while len(msg) < max_cols:
        msg += " "
    while len(msg) > max_cols:
        msg = msg[:-1]
    msg += "..."
    return msg



import logging
import os
import tarfile
from datetime import datetime
from pathlib import Path

from packaging.version import parse as parse_version
from tqdm import tqdm

from deemon import __version__
from deemon.utils import startup, dates

logger = logging.getLogger(__name__)


def run(include_logs: bool = False):
    def filter_func(item):
        includes = ['deemon', 'deemon/config.json', 'deemon/deemon.db']
        if include_logs:
            if 'deemon/logs' in item.name:
                includes.append(item.name)
        if item.name in includes:
            return item

    backup_tar = dates.generate_date_filename("backup-" + __version__ + "-") + ".tar"
    backup_path = startup.get_backup_dir()

    with tarfile.open(backup_path / backup_tar, "w") as tar:
        tar.add(startup.get_appdata_dir(), arcname='deemon', filter=filter_func)
        logger.info(f"Backed up to {backup_path / backup_tar}")


def restore():
    restore_file_list = ['deemon/config.json', 'deemon/deemon.db']

    def inspect_tar(fn: Path) -> dict:
        fn_name = fn.name
        fn_name = fn_name.replace('.tar', '').split('-')
        if fn_name[0] == "backup" and len(fn_name) > 3:
            if check_tar_contents(fn):
                backup_appversion = '-'.join(fn_name[1:-2])
                if is_newer_backup(backup_appversion):
                    logger.debug(f"Backup found for newer version {backup_appversion} is not compatible!")
                    return
                backup_time = datetime.strptime(fn_name[-1], "%H%M%S")
                backup_date = datetime.strptime(fn_name[-2], "%Y%m%d")
                try:
                    friendly_time = datetime.strftime(backup_time, "%-I:%M:%S %p")
                except ValueError:
                    # Gotta keep Windows happy...
                    friendly_time = datetime.strftime(backup_time, "%#I:%M:%S %p")
                try:
                    friendly_date = datetime.strftime(backup_date, "%b %-d, %Y")
                except ValueError:
                    # Gotta keep Windows happy...
                    friendly_date = datetime.strftime(backup_date, "%b %#d, %Y")
                backup_info = {
                    'version': backup_appversion,
                    'date': friendly_date,
                    'time': friendly_time,
                    'age': fn_name[-2] + fn_name[-1],
                    'filename': fn
                }
                return backup_info
        else:
            return

    def check_tar_contents(archive: Path):
        tar = tarfile.open(archive)
        file_list = tar.getmembers()
        files = [x.name for x in file_list]
        if all(item in files for item in restore_file_list):
            return True
        logger.debug("Archive is invalid or corrupt: " + str(archive))

    def restore_tarfile(archive: dict):
        logger.debug("Restoring backup from `" + str(archive['filename'].name + "`"))
        extract_dir = startup.get_appdata_dir()
        tar = tarfile.open(archive['filename'])
        progress = tqdm(tar.getmembers(), ascii=" #",
                        bar_format='{desc}  [{bar}] {percentage:3.0f}%')
        for member in progress:
            if member.isreg():
                if member.name in restore_file_list:
                    member.name = os.path.basename(member.name)
                    logger.info(f"Restoring {member.name}...")
                    progress.set_description_str(f"Restoring {member.name}")
                    tar.extract(member, extract_dir)
                    logger.debug(f"Restored {member.name} to {extract_dir}")
            if member == tar.getmembers()[-1]:
                progress.set_description_str("Restore complete")

    def is_newer_backup(version: str):
        if parse_version(version) > parse_version(__version__):
            return True

    def display_backup_list(available_backups: list):
        print("deemon Backup Manager\n")
        for index, backup in enumerate(available_backups, start=1):
            print(f"{index}. {backup['date']} @ {backup['time']} (ver {backup['version']})")

        selected_backup = int
        while selected_backup not in range(len(available_backups)):
            selected_backup = input("\nSelect a backup to restore (or press Enter to exit): ")
            if selected_backup == "":
                return
            try:
                selected_backup = int(selected_backup)
                selected_backup -= 1
            except ValueError:
                logger.warning("Invalid entry. Enter a number corresponding to the backup you wish to restore.")
        print("")
        restore_tarfile(available_backups[selected_backup])

    backups = []
    backup_path = startup.get_backup_dir()
    file_list = [x for x in sorted(Path(backup_path).glob('*.tar'))]
    for backup in file_list:
        tar_files = inspect_tar(backup)
        if tar_files:
            backups.append(tar_files)
    if backups:
        backups = sorted(backups, key=lambda x: x['age'], reverse=True)
        display_backup_list(backups)
    else:
        logger.info("No backups available to restore")



import logging
import time
from datetime import datetime, date

logger = logging.getLogger(__name__)


def get_todays_date():
    now_ts = int(time.time())
    today_date = datetime.utcfromtimestamp(now_ts).strftime('%Y-%m-%d')
    return today_date


def generate_date_filename(prefix: str):
    return prefix + datetime.today().strftime('%Y%m%d-%H%M%S')


def get_max_release_date(days):
    day_in_secs = 86400
    input_days_in_secs = days * day_in_secs
    max_date_ts = int(time.time()) - input_days_in_secs
    max_date = datetime.utcfromtimestamp(max_date_ts).strftime('%Y-%m-%d')
    return max_date


def get_year(release_date: str):
    return datetime.strptime(release_date, '%Y-%m-%d').year


def format_date_string(d: str):
    date_string = datetime.strptime(d, "%Y-%m-%d")
    return datetime.strftime(date_string, "%Y-%m-%d")


def str_to_datetime_obj(d: str) -> datetime:
    return datetime.strptime(d, "%Y-%m-%d")


def get_friendly_date(d: int):
    input_date = datetime.fromtimestamp(d).date()
    input_time = datetime.fromtimestamp(d).time()
    today = date.today()
    delta = today - input_date
    if delta.days == 0:
        try:
            return f"{input_time.strftime('%-I:%M %p')}"
        except ValueError:
            # Gotta keep Windows happy...
            return f"{input_time.strftime('%#I:%M %p')}"
    elif delta.days == 1:
        try:
            return f"Yesterday, {input_time.strftime('%-I:%M %p')}"
        except ValueError:
            # Gotta keep Windows happy...
            return f"Yesterday, {input_time.strftime('%#I:%M %p')}"

    elif 1 < delta.days < 7:
        try:
            return input_date.strftime("%A, ") + input_time.strftime("%-I:%M %p")
        except ValueError:
            # Gotta keep Windows happy...
            return input_date.strftime("%A, ") + input_time.strftime("%#I:%M %p")
    else:
        try:
            return input_date.strftime("%Y-%m-%d - ") + input_time.strftime("%-I:%M %p")
        except ValueError:
            # Gotta keep Windows happy...
            return input_date.strftime("%Y-%m-%d - ") + input_time.strftime("%#I:%M %p")



import csv
import logging
import os
import time
from pathlib import Path
from typing import Union

from deemon.core.db import Database
from deemon.utils.dates import generate_date_filename

logger = logging.getLogger(__name__)


class Show:

    def __init__(self):
        self.db = Database()

    def monitoring(self, artist: bool = True, query: str = None, export_csv: bool = False,
                   save_path: Union[str, Path] = None, filter: str = None, hide_header: bool = False,
                   is_id: bool = False):

        def csv_output(line: str):
            if save_path:
                output_to_file.append(line)
            else:
                print(line)

        output_to_file = []

        if artist:
            if query:
                if is_id:
                    try:
                        query = int(query)
                    except ValueError:
                        return logger.error(f"Invalid Artist ID - {query}")
                    db_result = self.db.get_monitored_artist_by_id(query)
                else:
                    db_result = self.db.get_monitored_artist_by_name(query)
            else:
                db_result = self.db.get_all_monitored_artists()

            if not db_result:
                if query:
                    return logger.error("Artist/ID not found: " + str(query))
                else:
                    return logger.error("No artists are being monitored")
        else:
            if query:
                if is_id:
                    try:
                        query = int(query)
                    except ValueError:
                        return logger.error(f"Invalid Playlist ID - {query}")
                    db_result = self.db.get_monitored_playlist_by_id(query)
                else:
                    db_result = self.db.get_monitored_playlist_by_name(query)
            else:
                db_result = self.db.get_all_monitored_playlists()

            if not db_result:
                if query:
                    return logger.error("Playlist/ID not found: " + str(query))
                else:
                    return logger.error("No playlists are being monitored")

        if artist and query:
            for key, val in db_result.items():
                if val == None:
                    db_result[key] = "-"

            print("{:<10} {:<35} {:<10} {:<10} {:<10} {:<25}".format('ID', 'Artist', 'Alerts',
                                                                     'Bitrate', 'Type', 'Download Path'))

            print("{!s:<10} {!s:<35} {!s:<10} {!s:<10} {!s:<10} {!s:<25}".format(db_result['artist_id'],
                                                                                 db_result['artist_name'],
                                                                                 db_result['alerts'],
                                                                                 db_result['bitrate'],
                                                                                 db_result['record_type'],
                                                                                 db_result['download_path']))
            print("")
        elif not artist and query:
            for key, val in db_result.items():
                if val == None:
                    db_result[key] = "-"

            print("{:<15} {:<30} {:<50} {:<10} {:<10} {:<25}".format('ID', 'Title', 'URL', 'Alerts',
                                                                     'Bitrate', 'Download Path'))

            print("{!s:<15} {!s:<30} {!s:<50}  {!s:<10} {!s:<10} {!s:<25}".format(db_result['id'], db_result['title'],
                                                                                  db_result['url'], db_result['alerts'],
                                                                                  db_result['bitrate'],
                                                                                  db_result['download_path']))
            print("")
        else:
            if export_csv or save_path:
                if artist:
                    if not filter:
                        filter = "name,id,bitrate,alerts,type,path"
                    filter = filter.split(',')
                    logger.debug(f"Generating CSV data using filters: {', '.join(filter)}")
                    column_names = ['artist_id' if x == 'id' else x for x in filter]
                    column_names = ['artist_name' if x == 'name' else x for x in column_names]
                    column_names = ['record_type' if x == 'type' else x for x in column_names]
                    column_names = ['download_path' if x == 'path' else x for x in column_names]
                else:
                    if not filter:
                        filter = "id,title,url,bitrate,alerts,path"
                    filter = filter.split(',')
                    logger.debug(f"Generating CSV data using filters: {', '.join(filter)}")
                    column_names = ['download_path' if x == 'path' else x for x in filter]

                for column in column_names:
                    if not len([x for x in db_result if column in x.keys()]):
                        logger.warning(f"Unknown filter specified: {column}")
                        column_names.remove(column)

                if not hide_header:
                    csv_output(','.join(filter))
                for artist in db_result:
                    filtered_artists = []
                    for key, value in artist.items():
                        if value is None:
                            artist[key] = ""
                    for column in column_names:
                        filtered_artists.append(str(artist[column]))
                    if len(filtered_artists):
                        for i, a in enumerate(filtered_artists):
                            if '"' in a:
                                a = a.replace('"', "'")
                            if ',' in a:
                                filtered_artists[i] = f'"{a}"'
                        csv_output(",".join(filtered_artists))

                if output_to_file:
                    if Path(save_path).is_dir():
                        output_filename = Path(save_path / f"{generate_date_filename('deemon_')}.csv")
                    else:
                        output_filename = Path(save_path)

                    with open(output_filename, 'w', encoding="utf-8") as f:
                        for line in output_to_file:
                            if line == output_to_file[-1]:
                                f.writelines(line)
                                break
                            f.writelines(line + "\n")

                    return logger.info("CSV data has been saved to: " + str(output_filename))

                return

            if artist:
                db_result = [x['artist_name'] for x in db_result]
            else:
                db_result = [x['title'] for x in db_result]
            if len(db_result) < 10:
                for artist in db_result:
                    print(artist)
            else:
                db_result = self.truncate_long_artists(db_result)

                size = os.get_terminal_size()
                max_cols = (int(size.columns / 30))
                if max_cols > 5:
                    max_cols = 5

                while len(db_result) % max_cols != 0:
                    db_result.append(" ")

                if max_cols >= 5:
                    for a, b, c, d, e in zip(db_result[0::5], db_result[1::5], db_result[2::5], db_result[3::5],
                                             db_result[4::5]):
                        print('{:<30}{:<30}{:<30}{:<30}{:<30}'.format(a, b, c, d, e))
                elif max_cols >= 4:
                    for a, b, c, d in zip(db_result[0::4], db_result[1::4], db_result[2::4], db_result[3::4]):
                        print('{:<30}{:<30}{:<30}{:<30}'.format(a, b, c, d))
                elif max_cols >= 3:
                    for a, b, c in zip(db_result[0::3], db_result[1::3], db_result[2::3]):
                        print('{:<30}{:<30}{:<30}'.format(a, b, c))
                elif max_cols >= 2:
                    for a, b in zip(db_result[0::2], db_result[1::2]):
                        print('{:<30}{:<30}'.format(a, b))
                else:
                    for a in db_result:
                        print(a)

    def playlists(self, csv=False):
        monitored_playlists = self.db.get_all_monitored_playlists()
        for p in monitored_playlists:
            print(f"{p[1]} ({p[2]})")

    @staticmethod
    def truncate_long_artists(all_artists):
        for idx, artist in enumerate(all_artists):
            if len(artist) > 25:
                all_artists[idx] = artist[:22] + "..."
            all_artists[idx] = artist
        return all_artists

    def releases(self, days, future):
        if future:
            future_releases = self.db.get_future_releases()
            future_release_list = [x for x in future_releases]
            if len(future_release_list) > 0:
                logger.info(f"Future releases:")
                print("")
                future_release_list.sort(key=lambda x: x['album_release'], reverse=True)
                for release in future_release_list:
                    print('+ [%-10s] %s - %s' % (release['album_release'], release['artist_name'], release['album_name']))
            else:
                logger.info("No future releases have been detected")
        else:
            seconds_per_day = 86400
            days_in_seconds = (days * seconds_per_day)
            now = int(time.time())
            back_date = (now - days_in_seconds)
            releases = self.db.show_new_releases(back_date, now)
            release_list = [x for x in releases]
            if len(release_list) > 0:
                logger.info(f"New releases found within last {days} day(s):")
                print("")
                release_list.sort(key=lambda x: x['album_release'], reverse=True)
                for release in release_list:
                    print('+ [%-10s] %s - %s' % (release['album_release'], release['artist_name'], release['album_name']))
            else:
                logger.info(f"No releases found in the last {days} day(s)")



import logging
import sys

from deezer import Deezer

from deemon.cmd import download
from deemon.cmd import monitor as mon
from deemon.core import db
from deemon.core.config import Config as config
from deemon.utils import dates

logger = logging.getLogger(__name__)


class Search:
    def __init__(self):
        self.artist_id: int = None
        self.artist: str = None
        self.choices: list = []
        self.status_message: str = None
        self.queue_list = []
        self.select_mode = False
        self.explicit_only = False
        self.user_search_query: str = None

        self.sort: str = "release_date"
        self.filter: str = None
        self.desc: bool = True

        self.db = db.Database()
        self.dz = Deezer()

    @staticmethod
    def truncate_artist(name: str):
        if len(name) > 45:
            return name[0:40] + "..."
        return name

    def get_latest_release(self, artist_id: int):
        try:
            all_releases = self.dz.api.get_artist_albums(artist_id)['data']
            sorted_releases = sorted(all_releases, key=lambda x: x['release_date'], reverse=True)
            latest_release = sorted_releases[0]
        except IndexError:
            return "       - No releases found"
        return f"       - Latest release: {latest_release['title']} ({dates.get_year(latest_release['release_date'])})"

    def display_monitored_status(self, artist_id: int):
        if self.db.get_monitored_artist_by_id(artist_id):
            return "[M] "
        return "    "

    @staticmethod
    def has_duplicate_artists(name: str, artist_dicts: dict):
        names = [x['name'] for x in artist_dicts if x['name'] == name]
        if len(names) > 1:
            return True

    def show_mini_queue(self):
        num_queued = len(self.queue_list)
        if num_queued > 0:
            return f" ({str(num_queued)} Queued)"
        return ""

    def search_menu(self, query: str = None):
        exit_search: bool = False
        quick_search: bool = False
        while exit_search is False:
            self.clear()
            print("deemon Interactive Search Client\n")
            if len(self.queue_list) > 0:
                self.display_options(options="(d) Download Queue  (Q) Show Queue")
            if query:
                search_query = query
                quick_search = True
            else:
                search_query = input(f":: Enter an artist to search for{self.show_mini_queue()}: ")
                if search_query == "exit":
                    if self.exit_search():
                        sys.exit()
                    continue
                if search_query == "d":
                    if len(self.queue_list) > 0:
                        self.start_queue()
                        continue
                if search_query == "Q":
                    if len(self.queue_list) > 0:
                        self.queue_menu()
                    else:
                        self.status_message = "Queue is empty"
                    continue
                if search_query == "":
                    continue
            artist_search_result = self.dz.api.search_artist(search_query, limit=config.query_limit())['data']
            if len(artist_search_result) == 0:
                self.status_message = "No results found for: " + search_query
                continue

            artist_selected = self.artist_menu(search_query, artist_search_result, quick_search)
            if artist_selected:
                self.user_search_query = search_query
                return [artist_selected]
            elif quick_search:
                return

    def queue_menu_options(self):
        ui_options = ("(d) Download Queue  (c) Clear Queue  (b) Back")
        self.display_options(options=ui_options)

    def artist_menu(self, query: str, results: dict, artist_only: bool = False):
        exit_artist: bool = False
        while exit_artist is False:
            self.clear()
            print("Search results for artist: " + query)
            for idx, option in enumerate(results, start=1):
                print(f"{self.display_monitored_status(option['id'])}{idx}. {self.truncate_artist(option['name'])}")
                if self.has_duplicate_artists(option['name'], results):
                    print(self.get_latest_release(option['id']))
                    print("       - Artist ID: " + str(option['id']))
                    if not option.get('nb_album'):
                        option['nb_album'] = self.dz.api.get_artist(option['id'])['nb_album']
                    print("       - Total releases: " + str(option['nb_album']))
                    self.status_message = "Duplicate artists found"
            # TODO make options smarter/modular
            if len(self.queue_list) > 0:
                self.display_options(options="(b) Back  (d) Download Queue  (Q) Show Queue")
            else:
                self.display_options(options="(b) Back")
            response = input(f":: Please choose an option or type 'exit'{self.show_mini_queue()}: ")
            if response == "d":
                if len(self.queue_list) > 0:
                    self.start_queue()
                    continue
            elif response == "Q":
                if len(self.queue_list) > 0:
                    self.queue_menu()
                else:
                    self.status_message = "Queue is empty"
                continue
            elif response == "b":
                break
            elif response == "exit":
                if self.exit_search() and not artist_only:
                    sys.exit()
                else:
                    return
            elif response == "":
                continue

            try:
                response = int(response)
            except ValueError:
                self.status_message = f"Invalid selection: {response}"
            else:
                response = response - 1
                if response in range(len(results)):
                    self.artist = results[response]['name']
                    if artist_only:
                        self.clear()
                        return results[response]
                    self.album_menu(results[response])
                else:
                    self.status_message = f"Invalid selection: {response}"
                    continue

    def album_menu_header(self, artist: str):
        filter_text = "All" if not self.filter else self.filter.title()
        if self.explicit_only:
            filter_text = filter_text + " (Explicit Only)"
        desc_text = "desc" if self.desc else "asc"
        sort_text = self.sort.replace("_", " ").title() + " (" + desc_text + ")"
        print("Discography for artist: " + artist)
        print("Filter by: " + filter_text + " | Sort by: " + sort_text + "\n")

    def album_menu_options(self, monitored):
        print("")
        if not monitored:
            monitor_opt = "(m) Monitor"
        else:
            monitor_opt = "(m) Stop Monitoring"
        ui_filter = "Filters: (*) All  (a) Albums  (e) EP  (s) Singles - (E) Explicit (r) Reset"
        ui_sort = "   Sort: (y) Year Desc  (Y) Year Asc  (t) Title Desc  (T) Title Asc"
        ui_mode = "   Mode: (S) Toggle Select"
        ui_options = ("(b) Back  (d) Download Queue  (Q) Show Queue  (f) Queue Filtered  "
                      f"{monitor_opt}")
        self.display_options(ui_filter, ui_sort, ui_mode, ui_options)

    @staticmethod
    def explicit_lyrics(is_explicit):
        if is_explicit:
            return " [E]"
        else:
            return ""

    def item_selected(self, id):
        if self.select_mode:
            if [x for x in self.queue_list if x.album_id == id or x.track_id == id]:
                return "[*] "
            else:
                return "[ ] "
        else:
            return "    "

    def show_mode(self):
        if self.select_mode:
            return "[SELECT] "
        return ""

    def album_menu(self, artist: dict):
        exit_album_menu: bool = False
        artist_albums = self.dz.api.get_artist_albums(artist['id'])['data']
        while exit_album_menu is False:
            self.clear()
            self.album_menu_header(artist['name'])
            filtered_choices = self.filter_choices(artist_albums)
            for idx, album in enumerate(filtered_choices, start=1):
                print(f"{self.item_selected(album['id'])}{idx}. ({dates.get_year(album['release_date'])}) "
                      f"{album['title']} {self.explicit_lyrics(album['explicit_lyrics'])}")
            monitored = self.db.get_monitored_artist_by_id(artist['id'])
            self.album_menu_options(monitored)

            prompt = input(f":: {self.show_mode()}Please choose an option or type 'exit'{self.show_mini_queue()}: ")
            if prompt == "a":
                self.filter = "album"
            elif prompt == "e":
                self.filter = "ep"
            elif prompt == "s":
                self.filter = "single"
            elif prompt == "*":
                self.filter = None
            elif prompt == "E":
                self.explicit_only ^= True
            elif prompt == "r":
                self.filter = None
                self.explicit_only = False
                self.sort = "release_date"
                self.desc = True
            elif prompt == "y":
                self.sort = "release_date"
                self.desc = True
            elif prompt == "Y":
                self.sort = "release_date"
                self.desc = False
            elif prompt == "t":
                self.sort = "title"
                self.desc = True
            elif prompt == "T":
                self.sort = "title"
                self.desc = False
            elif prompt == "S":
                self.select_mode ^= True
            elif prompt == "m":
                if monitored:
                    stop = True
                else:
                    stop = False
                record_type = self.filter or config.record_type()
                self.clear()
                monitor = mon.Monitor()
                monitor.set_config(None, None, record_type, None)
                monitor.set_options(stop, False, False)
                monitor.artist_ids([artist['id']])
            elif prompt == "f":
                if len(filtered_choices) > 0:
                    for item in filtered_choices:
                        self.send_to_queue(item)
                else:
                    self.status_message = "No items to add"
            elif prompt == "d":
                if len(self.queue_list) > 0:
                    self.start_queue()
            elif prompt == "Q":
                if len(self.queue_list) > 0:
                    self.queue_menu()
                else:
                    self.status_message = "Queue is empty"
            elif prompt == "b":
                break
            elif prompt == "":
                self.status_message = "Hint: to exit, type 'exit'!"
                continue
            elif prompt == "exit":
                if self.exit_search():
                    sys.exit()
            else:
                try:
                    selected_index = (int(prompt) - 1)
                except ValueError:
                    self.status_message = "Invalid filter, sort or option provided"
                    continue
                except IndexError:
                    self.status_message = "Invalid selection, please choose from above"
                    continue

                if selected_index in range(len(filtered_choices)):
                    if self.select_mode:
                        selected_item = filtered_choices[selected_index]
                        self.send_to_queue(selected_item)
                        continue
                    else:
                        self.track_menu(filtered_choices[selected_index])
                else:
                    self.status_message = "Invalid selection, please choose from above"
                    continue

    def track_menu_options(self):
        ui_options = ("(b) Back  (d) Download Queue  (Q) Show Queue")
        self.display_options(options=ui_options)

    def track_menu_header(self, album):
        print("deemon Interactive Search Client")
        print(f"Artist: {self.artist}  |  Album: {album['title']}\n")

    def track_menu(self, album):
        exit_track_menu: bool = False
        track_list = self.dz.api.get_album_tracks(album['id'])['data']
        self.select_mode = True
        while not exit_track_menu:
            self.clear()
            self.track_menu_header(album)

            for idx, track in enumerate(track_list, start=1):
                print(f"{self.item_selected(track['id'])}{idx}. {track['title']}")
            self.track_menu_options()

            prompt = input(f":: {self.show_mode()}Please choose an option or type 'exit'{self.show_mini_queue()}: ")
            if prompt == "d":
                if len(self.queue_list) > 0:
                    self.start_queue()
                else:
                    self.status_message = "Queue is empty"
            elif prompt == "Q":
                if len(self.queue_list) > 0:
                    self.queue_menu()
                else:
                    self.status_message = "Queue is empty"
            elif prompt == "b":
                self.select_mode = False
                break
            elif prompt == "":
                self.status_message = "Hint: to exit, type 'exit'!"
                continue
            elif prompt == "exit":
                if self.exit_search():
                    sys.exit()
            else:
                try:
                    selected_index = (int(prompt) - 1)
                except ValueError:
                    self.status_message = "Invalid filter, sort or option provided"
                    continue
                except IndexError:
                    self.status_message = "Invalid selection, please choose from above"
                    continue

                if selected_index in range(len(track_list)):
                    selected_item = track_list[selected_index]
                    self.send_to_queue(selected_item)
                    continue
                else:
                    self.status_message = "Invalid selection, please choose from above"
                    continue

    def search_header(self):
        pass

    def queue_menu(self):
        exit_queue_list = False
        while exit_queue_list is False:
            self.clear()
            for idx, q in enumerate(self.queue_list, start=1):
                if q.album_title:
                    print(f"{idx}. {q.artist_name} - {q.album_title}")
                else:
                    print(f"{idx}. {q.artist_name} - {q.track_title}")
            print("")
            self.queue_menu_options()
            response = input(f":: Please choose an option or type exit {self.show_mini_queue()}: ")
            if response == "d":
                if len(self.queue_list) > 0:
                    self.start_queue()
                else:
                    self.status_message = "Queue is empty"
            if response == "c":
                self.queue_list = []
                break
            if response == "b":
                break
            if response == "exit":
                if self.exit_search():
                    sys.exit()
            try:
                response = int(response) - 1
            except ValueError:
                continue
            if response in range(len(self.queue_list)):
                self.queue_list.pop(response)
                if len(self.queue_list) == 0:
                    break

    def exit_search(self):
        if len(self.queue_list) > 0:
            exit_all = input(":: Quit before downloading queue? [y|N] ")
            if exit_all.lower() != 'y':
                return False
        return True

    def display_options(self, filter=None, sort=None, mode=None, options=None):
        if filter:
            print(filter)
        if sort:
            print(sort)
        if mode:
            print(mode)
        if options:
            print("")
            print(options)
        if self.status_message:
            print("** " + self.status_message + " **")
            self.status_message = None

    @staticmethod
    def clear():
        from os import system, name
        if name == 'nt':
            _ = system('cls')
        else:
            _ = system('clear')

    def filter_choices(self, choices):
        apply_filter = [x for x in choices if x['record_type'] == self.filter or self.filter is None]
        if self.explicit_only:
            apply_filter = [x for x in apply_filter if x['explicit_lyrics'] == True]
        return sorted(apply_filter, key=lambda x: x[self.sort], reverse=self.desc)

    def start_queue(self):
        self.clear()
        dl = download.Download()
        dl.queue_list = self.queue_list
        dl.download_queue()
        self.queue_list.clear()
        self.status_message = "Downloads complete"

    def send_to_queue(self, item):
        if item['type'] == 'album':
            album = {'id': item['id'], 'title': item['title'], 'link': item['link'], 'artist': {'name': self.artist}}
            for i, q in enumerate(self.queue_list):
                if q.album_id == album['id']:
                    del self.queue_list[i]
                    return
            self.queue_list.append(download.QueueItem(album=album))

        elif item['type'] == 'track':
            track = {'id': item['id'], 'title': item['title'], 'link': item['link'], 'artist': self.artist}
            for i, q in enumerate(self.queue_list):
                if q.track_id == track['id']:
                    del self.queue_list[i]
                    return
            self.queue_list.append(download.QueueItem(track=track))

        elif item.get('name'):
            pass



import logging
from pathlib import Path

import deemix
import deemix.utils.localpaths as localpaths
from deemix import generateDownloadObject
from deemix.downloader import Downloader
from deemix.settings import load as LoadSettings
from deemix.types.DownloadObjects import Collection
from deezer import Deezer
from deezer.api import APIError
from deezer.gw import GWAPIError, LyricsStatus
from deezer.utils import map_user_playlist

from deemon.core import notifier
from deemon.core.config import Config as config
from deemon.core.db import Database

logger = logging.getLogger(__name__)


class DeemixInterface:
    def __init__(self):
        logger.debug("Initializing deemix library")
        self.db = Database()
        self.dz = Deezer()

        if config.deemix_path() == "":
            self.config_dir = localpaths.getConfigFolder()
        else:
            self.config_dir = Path(config.deemix_path())

        self.dx_settings = LoadSettings(self.config_dir)

        if config.download_path() != "":
            # TODO is this necessary?
            self.download_path = Path(config.download_path())
            self.dx_settings['downloadLocation'] = str(self.download_path)

        logger.debug("deemix " + deemix.__version__)
        logger.debug(f"deemix Config Path: {self.config_dir}")
        logger.debug(f"deemix Download Path: {self.dx_settings['downloadLocation']}")

    def download_url(self, url, bitrate, download_path, override_deemix=True):
        if override_deemix:
            deemix.generatePlaylistItem = self.generatePlaylistItem

        if download_path and download_path != "":
            self.dx_settings['downloadLocation'] = download_path

        links = []
        for link in url:
            if ';' in link:
                for l in link.split(";"):
                    links.append(l)
            else:
                links.append(link)
        for link in links:
            download_object = generateDownloadObject(self.dz, link, bitrate)
            if isinstance(download_object, list):
                for obj in download_object:
                    Downloader(self.dz, obj, self.dx_settings).start()
            else:
                Downloader(self.dz, download_object, self.dx_settings).start()

    def verify_arl(self, arl):
        if not self.dz.login_via_arl(arl):
            print("FAILED")
            logger.debug(f"ARL Failed: {arl}")
            return False
        print("OK")
        logger.debug("ARL is valid")
        return True

    def login(self):
        failed_logins = 0
        logger.debug("Looking for ARL...")
        if config.arl():
            logger.debug("ARL found in deemon config")
            print("Found ARL in deemon config, checking... ", end="")
            if self.verify_arl(config.arl()):
                return True
            else:
                failed_logins += 1

        if self.config_dir.is_dir():
            if Path(self.config_dir / '.arl').is_file():
                with open(self.config_dir / '.arl', 'r') as f:
                    arl_from_file = f.readline().rstrip("\n")
                    logger.debug("ARL found in deemix config")
                    print("Found ARL in deemix .arl file, checking... ", end="")
                    if self.verify_arl(arl_from_file):
                        return True
                    else:
                        failed_logins += 1
            else:
                logger.error(f"ARL not found in {self.config_dir}")
                return False
        else:
            logger.error(f"ARL directory {self.config_dir} was not found")
            return False

        if failed_logins > 1:
            notification = notifier.Notify()
            notification.expired_arl()

    def generatePlaylistItem(self, dz, link_id, bitrate, playlistAPI=None, playlistTracksAPI=None):
        if not playlistAPI:
            if not str(link_id).isdecimal():
                raise InvalidID(f"https://deezer.com/playlist/{link_id}")
            # Get essential playlist info
            try:
                playlistAPI = dz.api.get_playlist(link_id)
            except APIError:
                playlistAPI = None
            # Fallback to gw api if the playlist is private
            if not playlistAPI:
                try:
                    userPlaylist = dz.gw.get_playlist_page(link_id)
                    playlistAPI = map_user_playlist(userPlaylist['DATA'])
                except GWAPIError as e:
                    raise GenerationError(f"https://deezer.com/playlist/{link_id}", str(e)) from e

            # Check if private playlist and owner
            if not playlistAPI.get('public', False) and playlistAPI['creator']['id'] != str(self.dz.current_user['id']):
                logger.warning("You can't download others private playlists.")
                raise NotYourPrivatePlaylist(f"https://deezer.com/playlist/{link_id}")

        if not playlistTracksAPI:
            playlistTracksAPI = dz.gw.get_playlist_tracks(link_id)
        playlistAPI['various_artist'] = dz.api.get_artist(5080)  # Useful for save as compilation

        totalSize = len(playlistTracksAPI)
        playlistAPI['nb_tracks'] = totalSize
        collection = []
        for pos, trackAPI in enumerate(playlistTracksAPI, start=1):
            # Check if release has been seen already and skip it
            vals = {'track_id': trackAPI['SNG_ID'], 'playlist_id': playlistAPI['id']}
            sql = "SELECT * FROM 'playlist_tracks' WHERE track_id = :track_id AND playlist_id = :playlist_id"
            result = self.db.query(sql, vals).fetchone()
            if result:
                continue
            if trackAPI.get('EXPLICIT_TRACK_CONTENT', {}).get('EXPLICIT_LYRICS_STATUS', LyricsStatus.UNKNOWN) in [
                LyricsStatus.EXPLICIT, LyricsStatus.PARTIALLY_EXPLICIT]:
                playlistAPI['explicit'] = True
            trackAPI['POSITION'] = pos
            trackAPI['SIZE'] = totalSize
            collection.append(trackAPI)

        if 'explicit' not in playlistAPI:
            playlistAPI['explicit'] = False

        return Collection({
            'type': 'playlist',
            'id': link_id,
            'bitrate': bitrate,
            'title': playlistAPI['title'],
            'artist': playlistAPI['creator']['name'],
            'cover': playlistAPI['picture_small'][:-24] + '/75x75-000000-80-0-0.jpg',
            'explicit': playlistAPI['explicit'],
            'size': totalSize,
            'collection': {
                'tracks_gw': collection,
                'playlistAPI': playlistAPI
            }
        })


class GenerationError(Exception):
    def __init__(self, link, message, errid=None):
        super().__init__()
        self.link = link
        self.message = message
        self.errid = errid

    def toDict(self):
        return {
            'link': self.link,
            'error': self.message,
            'errid': self.errid
        }


class InvalidID(GenerationError):
    def __init__(self, link):
        super().__init__(link, "Link ID is invalid!", "invalidID")


class NotYourPrivatePlaylist(GenerationError):
    def __init__(self, link):
        super().__init__(link, "You can't download others private playlists.", "notYourPrivatePlaylist")



import logging
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timedelta

from tqdm import tqdm

from deemon.cmd.download import QueueItem, Download
from deemon.core import db, api, notifier
from deemon.core.config import Config as config
from deemon.utils import dates, ui, performance

logger = logging.getLogger(__name__)


class Refresh:
    def __init__(self, time_machine: list = None, skip_download: bool = False, ignore_filters: bool = False):
        self.db = db.Database()
        self.refresh_date = datetime.now()
        self.max_refresh_date = None
        self.api = api.PlatformAPI("deezer-gw")
        self.new_releases = []
        self.new_releases_alert = []
        self.new_playlist_releases = []
        self.time_machine = False
        self.total_new_releases = 0
        self.queue_list = []
        self.skip_download = skip_download
        self.download_all = ignore_filters
        
        if not time_machine: time_machine = []
        if len(time_machine):
            if len(time_machine) == 2:
                self.refresh_date = time_machine[1]
                self.max_refresh_date = time_machine[0]
                logger.info(":: Time Machine active: "
                            f"{datetime.strftime(self.max_refresh_date, '%b %d, %Y')}"
                            f" - {datetime.strftime(self.refresh_date, '%b %d, %Y')}!")
            else:
                self.refresh_date = time_machine[0]
                logger.info(f":: Time Machine active: {datetime.strftime(self.refresh_date, '%b %d, %Y')}!")
            self.time_machine = True
            config.set('by_release_date', False)

    def debugger(self, message: str, payload = None):
        if config.debug_mode():
            if not payload:
                payload = ""
            logger.debug(f"DEBUG_MODE: {message} {str(payload)}")

    def remove_existing_releases(self, payload: dict) -> list:
        """
        Return list of releases that have not been stored in the database
        """
        new_releases = []

        if payload.get('artist_id'):
            seen_releases = self.db.get_artist_releases()
            if seen_releases:
                seen_releases = [v for x in seen_releases for k, v in x.items()]
                new_releases = [x for x in payload['releases'] if type(x) == dict for k, v in x.items() if
                                k == "id" and v not in seen_releases]
                return new_releases
            return [x for x in payload['releases']]

        if payload.get('tracks'):
            playlist_id = payload['id']
            seen_releases = self.db.get_playlist_tracks(playlist_id)
            if seen_releases:
                seen_releases = [v for x in seen_releases for k, v in x.items()]
                new_releases = [x for x in payload['tracks']
                                if type(x) == dict for k, v in x.items()
                                if k == "id" and v not in seen_releases]
                return new_releases
            return [x for x in payload['tracks']]

        return new_releases

    def filter_artist_releases(self, payload: dict):
        """ Inspect artist releases and decide what to do with each release """
        
        self.debugger("FilterReleases", {'artist': payload['artist_id'],
                                         'releases': len(payload['releases'])})
        for release in payload['releases']:
            release['artist_id'] = payload['artist_id']
            release['artist_name'] = payload['artist_name']
            release['future'] = self.is_future_release(release['release_date'])
            
            if release['explicit_lyrics'] != 1:
                release['explicit_lyrics'] = 0
            
            self.append_database_release(release)
            
            if release['future']:
                continue
            
            explicit_album_id = self.explicit_id(release['title'], payload['releases'])
            if explicit_album_id:
                if explicit_album_id == release['id']:
                    logger.debug(f"An explicit release was found for {release['title']}")
                else:
                    continue
            
            if self.download_all:
                self.queue_release(release)
                continue
            
            if not self.allowed_record_type(payload['record_type'],
                                            release['record_type']):
                logger.debug(f"Record type \"{release['record_type']}\" "
                                "has been filtered out, skipping release "
                                f"{release['id']}")
                continue
            
            if self.release_too_old(release['release_date']):
                logger.debug(f"Release {release['id']} is too old, "
                                "skipping it.")
                continue
            
            if not payload['refreshed']:
                if not self.time_machine:
                    continue
            
            self.queue_release(release)
            
    def append_database_release(self, new_release: dict):
        self.new_releases.append(new_release)
                
    @staticmethod
    def explicit_id(release_title: str, payload: list):
        explicit_found = {}
        for release in payload:
            if release['title'] == release_title:
                if release['explicit_lyrics'] == 1:
                    return release['id']
        
    
    def release_too_old(self, release_date: str):
        release_date_dt = dates.str_to_datetime_obj(release_date)
        if self.time_machine:
            if self.max_refresh_date:
                if release_date_dt < self.max_refresh_date:
                    return True
                if release_date_dt > self.refresh_date:
                    return True
            elif release_date_dt < self.refresh_date:
                return True
        elif config.release_by_date():
            if release_date_dt < (self.refresh_date - timedelta(config.release_max_age())):
                return True
                
                
    def is_future_release(self, release_date: str):
        """ Return 1 if release date is in future, otherwise return 0 """
        
        release_date_dt = dates.str_to_datetime_obj(release_date)
        if release_date_dt > datetime.now():
            return 1
        else:
            return 0
    
    def allowed_record_type(self, artist_rec_type, release_rec_type: str):
        """ Compare actual record_type against allowable """
        
        if artist_rec_type:
            if artist_rec_type == release_rec_type:
                return True
            else:
                return
        elif config.record_type() == release_rec_type:
            return True
        elif config.record_type() == "all":
            return True

    def queue_release(self, release: dict):
        """ Add release to download queue and create alert notification """
        
        self.create_notification(release)
        self.queue_list.append(QueueItem(release_full=release))

    def filter_playlist_releases(self, payload: dict):
        self.debugger(f"Filtering {len(payload['tracks'])} tracks for playlist {payload['title']}")
        if len(payload['tracks']):
            for track in payload['tracks']:
                new_track = track.copy()
                new_track['playlist_id'] = payload['id']
                self.new_playlist_releases.append(new_track)
                
                if payload['refreshed'] == 0:
                    continue
                
                queue_obj = QueueItem(playlist=payload, bitrate=payload['bitrate'],
                                        download_path=payload['download_path'])
                self.debugger("QueuePlaylistItem", queue_obj)
                self.queue_list.append(queue_obj)

    def waiting_for_refresh(self):
        playlists = self.db.get_unrefreshed_playlists()
        artists = self.db.get_unrefreshed_artists()
        if len(playlists) or len(artists):
            return {'artists': artists, 'playlists': playlists}

    # @performance.timeit
    def run(self, artists: list = None, playlists: list = None):
        if artists:
            self.debugger("ManualRefresh", artists)
            monitored_artists = [x for x in (self.db.get_monitored_artist_by_name(a) for a in artists) if x]
            if not len(monitored_artists):
                return logger.warning("Specified artist(s) were not found")
            api_result = self.get_release_data({'artists': monitored_artists})
        elif playlists:
            self.debugger("ManualRefresh", playlists)
            monitored_playlists = [x for x in (self.db.get_monitored_playlist_by_name(p) for p in playlists) if x]
            if not len(monitored_playlists):
                return logger.warning("Specified playlist(s) were not found")
            api_result = self.get_release_data({'playlists': monitored_playlists})
        else:
            waiting = self.waiting_for_refresh()
            if waiting:
                logger.debug(f"There are {len(waiting['playlists'])} playlist(s) and "
                             f"{len(waiting['artists'])} artist(s) waiting to be refreshed.")
                api_result = self.get_release_data(waiting)
            else:
                self.debugger("FullRefresh")
                monitored_playlists = self.db.get_all_monitored_playlists()
                monitored_artists = self.db.get_all_monitored_artists()
                if not len(monitored_playlists) and not len(monitored_artists):
                    return logger.warning("No artists found to refresh")
                api_result = self.get_release_data({'artists': monitored_artists, 'playlists': monitored_playlists})

        artist_processor = tqdm(api_result['artists'], total=len(api_result['artists']), desc="Checking for new releases...",
                                ascii=" #", bar_format=ui.TQDM_FORMAT)
        for payload in artist_processor:
            if len(payload):
                payload['releases'] = self.remove_existing_releases(payload)
                self.filter_artist_releases(payload)

        for payload in api_result['playlists']:
            if len(payload):
                payload['tracks'] = self.remove_existing_releases(payload)
                self.filter_playlist_releases(payload)

        if self.skip_download:
            logger.info(f"You have opted to skip downloads, emptying {len(self.queue_list)} item(s) from queue...")
            self.queue_list.clear()
            self.new_releases_alert.clear()

        if len(self.queue_list):
            dl = Download()
            dl.download_queue(self.queue_list)

        if len(self.new_playlist_releases) or len(self.new_releases):
            if len(self.new_playlist_releases):
                logger.debug("Updating playlist releases in database...")
                self.db.add_new_playlist_releases(self.new_playlist_releases)
            if len(self.new_releases):
                logger.debug("Updating artist releases in database...")
                self.db.add_new_releases(self.new_releases)
            self.db.commit()
            self.db_stats()
            performance.operation_time(config.get('start_time'))
            logger.info("Database is up-to-date.")
        else:
            self.db_stats()
            performance.operation_time(config.get('start_time'))
            logger.info("Database is up-to-date. No new releases were found.")

        if len(self.new_releases_alert) > 0:
            notification = notifier.Notify(self.new_releases_alert)
            notification.send()

    def db_stats(self):
        artists = len(self.db.get_all_monitored_artist_ids())
        playlists = len(self.db.get_all_monitored_playlist_ids())
        releases = len(self.db.get_artist_releases())
        future = len(self.db.get_future_releases())

        print("")
        print(f"+ Artists monitored: {artists:,}")
        print(f"+ Playlists monitored: {playlists:,}")
        print(f"+ Releases seen: {releases:,}")
        print(f"+ Pending future releases: {future:,}")
        print("")

    # @performance.timeit
    def get_release_data(self, to_refresh: dict) -> dict:
        """
        Generate a list of dictionaries containing artist (DB) and release (API)
        information.
        """
        if self.time_machine:
            logger.debug("Time machine has been detected; clearing future releases...")
            artist_ids = [{'id': artist['artist_id']} for artist in to_refresh['artists']]
            playlist_ids = [{'id': playlist['id']} for playlist in to_refresh['playlists']]
            self.db.remove_specific_releases(artist_ids)
            self.db.remove_specific_playlist_tracks(playlist_ids)

        api_result = {'artists': [], 'playlists': []}

        logger.debug(f"Standby, starting refresh...")

        if to_refresh.get('playlists') and len(to_refresh.get('playlists')):
            logger.debug("Fetching playlist track data...")
            self.debugger("SpawningThreads", self.api.max_threads)
            with ThreadPoolExecutor(max_workers=self.api.max_threads) as ex:
                api_result['playlists'] = list(
                    tqdm(ex.map(self.api.get_playlist_tracks,
                                to_refresh['playlists']),
                         total=len(to_refresh['playlists']),
                         desc=f"Fetching playlist track data for "
                              f"{len(to_refresh['playlists'])} playlist(s), "
                              "please wait...",
                         ascii=" #",
                         bar_format=ui.TQDM_FORMAT)
                )

        if to_refresh.get('artists') and len(to_refresh['artists']):
            logger.debug("Fetching artist release data...")
            self.debugger("SpawningThreads", self.api.max_threads)
            with ThreadPoolExecutor(max_workers=self.api.max_threads) as ex:
                api_result['artists'] = list(
                    tqdm(ex.map(self.api.get_artist_albums, to_refresh['artists']),
                         total=len(to_refresh['artists']), desc=f"Fetching artist release data for {len(to_refresh['artists']):,} artist(s), please wait...", ascii=" #",
                         bar_format=ui.TQDM_FORMAT)
                )
        return api_result

    def create_notification(self, release: dict):
        for days in self.new_releases_alert:
            for key in days:
                if key == "release_date":
                    if release['release_date'] in days[key]:
                        days["releases"].append({'artist': release['artist_name'], 'album': release['title']})
                        return

        self.new_releases_alert.append({'release_date': release['release_date'], 'releases': [{'artist': release['artist_name'], 'album': release['title']}]})



from pathlib import Path

import tqdm as tqdm
from deezer import Deezer


def read_album_ids_from_file(filename):
    if not Path(filename).exists():
        raise Exception("File does not exist")
    ids = []
    with open(filename, 'r') as f:
        f.readline()
        for l in f:
            ids.append(l.encode("ascii", "ignore"))
        print("Total lines read from text file: " + str(len(ids)))
        return ids


def clean_absolute_paths(album_list):
    stripped = []
    for line in album_list:
        stripped.append(line.split("\\"))
    return stripped


def clean_year_from_album(album_list, level: int = 5):
    artist_album = []
    for line in album_list:
        if len(line) == level:
            strip_year = line[3][-6:].strip("()")
            try:
                int(strip_year)
            except Exception:
                artist_album.append([line[2], line[3]])
                continue
            artist_album.append([line[2], line[3][:-6]])
    return artist_album


def clean_artist_album_text(album_list: list):
    stripped = []
    for line in album_list:
        line = line.decode()
        line = line.replace('\n', '')
        line = line.replace('�', '')
        line = line.replace('¡', '')
        line = line.replace('É', 'E')
        line = line.replace('.', '')
        line = line.replace('+', ' ')
        line = line.replace('/', ' ')
        stripped.append(line.split(" - "))
    return stripped


def get_artist_album(filename: str, absolute_path: bool = False):
    list_from_file = read_album_ids_from_file(filename)
    if absolute_path:
        stripped_paths = clean_absolute_paths(list_from_file)
        artist_album = clean_year_from_album(stripped_paths, level=5)
    else:
        artist_album = clean_artist_album_text(list_from_file)
    print("Total albums to lookup: " + str(len(artist_album)))
    return sorted(x for x in artist_album)


def get_api_results(album_list, artist_name: str = None):
    dz = Deezer()

    for x in album_list:
        album_list.set_description_str(f"Pass: {str(len(id_list))} | Fail: {str(len(fail_list))}")
        # For testing, only process this artist
        if artist_name and artist_name != x[0]:
            continue
        # Remove things like "(Bonus Tracks)"
        artist_from_file = x[0]
        album_from_file = x[1]
        stripped_album_from_file = x[1].split(" (")[0]

        api_artist = dz.api.search_artist(x[0], limit=10)['data']
        found_artist = True
        for artist_result in api_artist:
            # TODO name decode - replace unknown with ? - use as wildcard
            encoded_name = artist_result['name'].encode("ascii", "replace")
            decoded_name = encoded_name.decode()
            if artist_from_file == decoded_name:
                api_artist = artist_result
                found_artist = True
                break
            else:
                found_artist = False

        # TODO make this add albums to id and break out
        if found_artist is False:
            for artist in api_artist:
                get_albums = dz.api.get_artist_albums(artist['id'])['data']
                if album_from_file in [x['title'] for x in get_albums]:
                    api_artist = artist
                    break
            #
            # print("Searched all albums, nothing matches!")
            # exit()
        try:
            all_albums = dz.api.get_artist_albums(api_artist['id'])['data']
        except:
            fail_list.append(f"{x[0]} - {x[1]}")
            continue
        api_albums = [[x['title'].split(" (")[0], x['id']] for x in all_albums]
        for [title, id] in api_albums:
            clean_title = title
            clean_title = clean_title.replace('¡', '')
            clean_title = clean_title.replace('É', 'E')
            clean_title = clean_title.replace('.', '')
            clean_title = clean_title.replace(' + ', ' ')
            clean_title = clean_title.replace(' / ', ' ')
            if album_from_file.lower() == clean_title.lower() or stripped_album_from_file.lower() == clean_title.lower():
                if id not in id_list:
                    id_list.append(id)
                break
            if album_from_file.lower() in clean_title.lower() or stripped_album_from_file.lower() in clean_title.lower():
                if id not in id_list:
                    id_list.append(id)
                break
        else:
            fail_list.append(f"{x[0]} - {x[1]}")


id_list = []
fail_list = []
input_file_or_directory: str = None
output_file_passing: str = None
output_file_failing: str = None

album_list = get_artist_album(input_file_or_directory)

progress = tqdm.tqdm(album_list, ascii=" #",
                     bar_format='{desc}...  {n_fmt}/{total_fmt} [{bar:40}] {percentage:3.0f}%')
progress.set_description_str("Getting IDs")

get_api_results(progress)

print("PASS: " + str(len(id_list)))
with open(output_file_passing, "w") as f:
    for id in id_list:
        f.write(str(id) + "\n")

print("FAIL: " + str(len(fail_list)))
with open(output_file_failing, "w") as f:
    for id in fail_list:
        f.write(str(id) + "\n")



import logging

from deemon.core.config import Config as config
from deemon.core.db import Database

logger = logging.getLogger(__name__)


class ProfileConfig:
    def __init__(self, profile_name):
        self.db = Database()
        self.profile_name = profile_name
        self.profile = None

    # TODO move this to utils
    @staticmethod
    def print_header(message: str = None):
        print("deemon Profile Editor")
        if message:
            print(":: " + message + "\n")
        else:
            print("")

    def edit(self):
        profile = self.db.get_profile(self.profile_name)
        self.print_header(f"Configuring '{profile['name']}' (Profile ID: {profile['id']})")
        modified = 0
        for property in profile:
            if property == "id":
                continue
            allowed_opts = config.allowed_values(property)
            if isinstance(allowed_opts, dict):
                allowed_opts = [str(x.lower()) for x in allowed_opts.values()]

            while True:
                friendly_text = property.replace("_", " ").title()
                user_input = input(f"{friendly_text} [{profile[property]}]: ").lower()
                if user_input == "":
                    break
                # TODO move to function to share with Config.set()?
                elif user_input == "false" or user_input == "0":
                    user_input = False
                elif user_input == "true" or user_input == "1":
                    user_input = True
                elif property == "name" and self.profile_name != user_input:
                    if self.db.get_profile(user_input):
                        print("Name already in use")
                        continue
                if user_input == "none" and property != "name":
                    user_input = None
                elif allowed_opts:
                    if user_input not in allowed_opts:
                        print(f"Allowed options: " + ', '.join(str(x) for x in allowed_opts))
                        continue
                logger.debug(f"User set {property} to {user_input}")
                profile[property] = user_input
                modified += 1
                break

        if modified > 0:
            user_input = input("\n:: Save these settings? [y|N] ")
            if user_input.lower() != "y":
                logger.info("No changes made, exiting...")
            else:
                self.db.update_profile(profile)
                print(f"\nProfile '{profile['name']}' has been updated!")
        else:
            print("No changes made, exiting...")

    def add(self):
        new_profile = {}
        profile_config = self.db.get_profile(self.profile_name)
        if profile_config:
            return logger.error(f"Profile {self.profile_name} already exists")
        else:
            logger.info("Adding new profile: " + self.profile_name)
            print("** Any option left blank will fallback to global config **\n")
            new_profile['name'] = self.profile_name

        menu = [
            {'setting': 'email', 'type': str, 'text': 'Email address', 'allowed': []},
            {'setting': 'alerts', 'type': bool, 'text': 'Alerts', 'allowed': config.allowed_values('alerts')},
            {'setting': 'bitrate', 'type': str, 'text': 'Bitrate',
             'allowed': config.allowed_values('bitrate').values()},
            {'setting': 'record_type', 'type': str, 'text': 'Record Type',
             'allowed': config.allowed_values('record_type')},
            {'setting': 'plex_baseurl', 'type': str, 'text': 'Plex Base URL', 'allowed': []},
            {'setting': 'plex_token', 'type': str, 'text': 'Plex Token', 'allowed': []},
            {'setting': 'plex_library', 'type': str, 'text': 'Plex Library', 'allowed': []},
            {'setting': 'download_path', 'type': str, 'text': 'Download Path', 'allowed': []},
        ]

        for m in menu:
            repeat = True
            while repeat:
                i = input(m['text'] + ": ")
                if i == "":
                    new_profile[m['setting']] = None
                    break
                if not isinstance(i, m['type']):
                    try:
                        i = int(i)
                    except ValueError:
                        print(" - Allowed options: " + ', '.join(str(x) for x in m['allowed']))
                        continue
                if len(m['allowed']) > 0:
                    if i not in m['allowed']:
                        print(" - Allowed options: " + ', '.join(str(x) for x in m['allowed']))
                        continue
                new_profile[m['setting']] = i
                break

        print("\n")
        i = input(":: Save these settings? [y|N] ")
        if i.lower() != "y":
            return logger.info("Operation cancelled. No changes saved.")
        else:
            self.db.create_profile(new_profile)
            logger.debug(f"New profile created with the following configuration: {new_profile}")

    def delete(self):
        profile_config = self.db.get_profile(self.profile_name)
        if not profile_config:
            return logger.error(f"Profile {self.profile_name} not found")

        if profile_config['id'] == 1:
            return logger.info("You cannot delete the default profile.")

        i = input(f":: Remove the profile '{self.profile_name}'? [y|N] ")
        if i.lower() == "y":
            self.db.delete_profile(self.profile_name)
            return logger.info("Profile " + self.profile_name + " deleted.")
        else:
            return logger.info("Operation cancelled")

    def show(self):
        if not self.profile_name:
            profile = self.db.get_all_profiles()
            self.print_header(f"Showing all profiles")
        else:
            profile = [self.db.get_profile(self.profile_name)]
            self.print_header(f"Showing profile '{profile[0]['name']}' (Profile ID: {profile[0]['id']})")
            if len(profile) == 0:
                return logger.error(f"Profile {self.profile_name} not found")

        print("{:<10} {:<40} {:<8} {:<8} {:<8} {:<25} "
              "{:<20} {:<20} {:<20}".format('Name', 'Email', 'Alerts', 'Bitrate', 'Type',
                                            'Plex Base URL', 'Plex Token', 'Plex Library', 'Download Path'))
        for u in profile:
            id, name, email, alerts, bitrate, rtype, url, token, \
            lib, dl_path = [x if x is not None else '' for x in u.values()]
            print("{:<10} {:<40} {:<8} {:<8} {:<8} {:<25} "
                  "{:<20} {:<20} {:<20}".format(name, email, alerts, bitrate, rtype, url, token, lib, dl_path))
            print("")

    def clear(self):
        profile = self.db.get_profile(self.profile_name)
        self.print_header(f"Configuring '{profile['name']}' (Profile ID: {profile['id']})")
        if not profile:
            return logger.error(f"Profile {self.profile_name} not found")

        for value in profile:
            if value in ["id", "name"]:
                continue
            profile[value] = None
        self.db.update_profile(profile)
        logger.info("All values have been cleared.")



import logging
import time

logger = logging.getLogger(__name__)


def timeit(method):
    def timed(*args, **kwargs):
        ts = time.time()
        result = method(*args, **kwargs)
        te = time.time()

        logger.debug(f"{method.__name__} finished in ({str((te - ts))})")
        return result

    return timed


def operation_time(start_time):
    end_time = int(time.time())
    duration = end_time - start_time
    output = time.strftime("%H:%M:%S", time.gmtime(duration))
    logger.info(f"Operation completed in {output}")


import logging
from logging.handlers import RotatingFileHandler

import tqdm

LOG_FORMATS = {
    'DEBUG': '%(asctime)s [%(levelname)s] %(name)s: %(message)s',
    'INFO': '%(message)s',
}

LOG_DATE = '%Y-%m-%d %H:%M:%S'


class tqdmStream(object):

    @classmethod
    def write(cls, msg):
        tqdm.tqdm.write(msg, end='')


def setup_logger(log_level='DEBUG', log_file=None):
    """
    Configure logging for the deemon application
    """

    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)

    deemon_logger = logging.getLogger("deemon")
    deemon_logger.setLevel(logging.DEBUG)

    deemix_logger = logging.getLogger("deemix")
    deemix_logger.setLevel(logging.DEBUG)

    urllib3_logger = logging.getLogger("urllib3")
    urllib3_logger.setLevel(logging.ERROR)

    spotipy_logger = logging.getLogger("spotipy")
    spotipy_logger.setLevel(logging.INFO)

    del logger.handlers[:]
    del deemix_logger.handlers[:]

    if log_file is not None:
        rotate = RotatingFileHandler(log_file, maxBytes=2097152, backupCount=5, encoding="utf-8")
        rotate.setLevel(logging.DEBUG)
        rotate.setFormatter(logging.Formatter(LOG_FORMATS['DEBUG'], datefmt=LOG_DATE))
        logger.addHandler(rotate)

    stream = logging.StreamHandler(stream=tqdmStream)
    stream.setLevel(log_level)
    stream.setFormatter(logging.Formatter(LOG_FORMATS[log_level], datefmt=LOG_DATE))
    deemon_logger.addHandler(stream)



import logging
import os
from pathlib import Path

import deemix.errors
import deezer
import plexapi.exceptions
from plexapi.server import PlexServer

from deemon import utils
from deemon.core import dmi
from deemon.core.config import Config as config
from deemon.utils import dataprocessor, validate, startup, dates

logger = logging.getLogger(__name__)


class QueueItem:
    # TODO - Accept new playlist tracks for output/alerts
    def __init__(self, artist=None, album=None, track=None, playlist=None,
                 bitrate: str = None, download_path: str = None,
                 release_full: dict = None):
        self.artist_name = None
        self.album_id = None
        self.album_title = None
        self.track_id = None
        self.track_title = None
        self.url = None
        self.playlist_title = None
        self.bitrate = bitrate or config.bitrate()
        self.download_path = download_path or config.download_path()
        
        if release_full:
            self.artist_name = release_full['artist_name']
            self.album_id = release_full['id']
            self.album_title = release_full['title']
            self.url = f"https://www.deezer.com/album/{self.album_id}"

        if artist:
            try:
                self.artist_name = artist["artist_name"]
            except KeyError:
                self.artist_name = artist["name"]
            if not album and not track:
                self.url = artist["link"]

        if album:
            if not artist:
                self.artist_name = album["artist"]["name"]
            self.album_id = album["id"]
            self.album_title = album["title"]
            try:
                self.url = album["link"]
            except KeyError:
                self.url = f"https://www.deezer.com/album/{album['id']}"

        if track:
            self.artist_name = track["artist"]
            self.track_id = track["id"]
            self.track_title = track["title"]
            self.url = track["link"]

        if playlist:
            self.url = playlist["url"]
            self.playlist_title = playlist["title"]


def get_deemix_bitrate(bitrate: str):
    for bitrate_id, bitrate_name in config.allowed_values('bitrate').items():
        if bitrate_name.lower() == bitrate.lower():
            return bitrate_id


def get_plex_server():
    if (config.plex_baseurl() != "") and (config.plex_token() != ""):
        try:
            print("Plex settings found, trying to connect (10s)... ", end="")
            plex_server = PlexServer(config.plex_baseurl(), config.plex_token(), timeout=10)
            print(" OK")
            return plex_server
        except Exception:
            print(" FAILED")
            logger.error("Error: Unable to reach Plex server, please refresh manually.")
            return False


def refresh_plex(plexobj):
    try:
        plexobj.library.section(config.plex_library()).update()
        logger.debug("Plex library refreshed successfully")
    except plexapi.exceptions.BadRequest as e:
        logger.error("Error occurred while refreshing your library. See logs for additional info.")
        logger.debug(f"Error during Plex refresh: {e}")
    except plexapi.exceptions.NotFound as e:
        logger.error("Error: Plex library not found. See logs for additional info.")
        logger.debug(f"Error during Plex refresh: {e}")


class Download:

    def __init__(self):
        super().__init__()
        self.dz = deezer.Deezer()
        self.di = dmi.DeemixInterface()
        self.queue_list = []
        self.bitrate = None
        self.from_release_date = None
        self.verbose = os.environ.get("VERBOSE")
        self.duplicate_id_count = 0

    # @performance.timeit
    def download_queue(self, queue_list: list = None):
        if queue_list:
            self.queue_list = queue_list

        if not self.di.login():
            return False

        if self.queue_list:
            plex = get_plex_server()
            print("----------------------------")
            logger.info("Sending " + str(len(self.queue_list)) + " release(s) to deemix for download:")

            with open(startup.get_appdata_dir() / "queue.csv", "w", encoding="utf-8") as f:
                f.writelines(','.join([str(x) for x in vars(self.queue_list[0]).keys()]) + "\n")
                logger.debug(f"Writing queue to CSV file - {len(self.queue_list)} items in queue")
                for q in self.queue_list:
                    raw_values = [str(x) for x in vars(q).values()]
                    # TODO move this to shared function
                    for i, v in enumerate(raw_values):
                        if '"' in v:
                            raw_values[i] = v.replace('"', "'")
                        if ',' in v:
                            raw_values[i] = f'"{v}"'
                    f.writelines(','.join(raw_values) + "\n")
            logger.debug(f"Queue exported to {startup.get_appdata_dir()}/queue.csv")

            def send_queue_to_deemix(queue) -> list:
                dx_bitrate = get_deemix_bitrate(queue.bitrate)
                if self.verbose == "true":
                    logger.debug(f"Processing queue item {vars(queue)}")
                try:
                    if queue.artist_name:
                        if queue.album_title:
                            logger.info(f"{queue.artist_name} - {queue.album_title}... ")
                            self.di.download_url([queue.url], dx_bitrate, config.download_path())
                        else:
                            logger.info(f"{queue.artist_name} - {queue.track_title}... ")
                            self.di.download_url([queue.url], dx_bitrate, config.download_path())
                    else:
                        logger.info(f"{queue.playlist_title} (playlist)...")
                        self.di.download_url([queue.url], dx_bitrate, queue.download_path, override_deemix=True)
                except deemix.errors.GenerationError:
                    return [(queue, "No tracks listed or unavailable in your country")]

            failed_count = []
            for q in self.queue_list:
                failed_count.append(send_queue_to_deemix(q))
            # with ThreadPoolExecutor(max_workers=5) as ex:
            #     failed_count = list(tqdm(ex.map(send_queue_to_deemix, self.queue_list),
            #                              total=len(self.queue_list), desc="Downloading        ...", ascii=" #",
            #                              bar_format='[{n_fmt}/{total_fmt}] {desc} [{bar}] {percentage:3.0f}%'))

            failed_count = [x for x in failed_count if x]

            print("")
            if len(failed_count):
                logger.info(f"Downloads completed with {len(failed_count)} error(s):")
                with open(startup.get_appdata_dir() / "failed.csv", "w", encoding="utf-8") as f:
                    f.writelines(','.join([str(x) for x in vars(self.queue_list[0]).keys()]) + "\n")
                    for failed in failed_count:
                        raw_values = [str(x) for x in vars(failed[0]).values()]
                        # TODO move this to shared function
                        for i, v in enumerate(raw_values):
                            if '"' in v:
                                raw_values[i] = v.replace('"', "'")
                            if ',' in v:
                                raw_values[i] = f'"{v}"'
                        f.writelines(','.join(raw_values) + "\n")
                        print(f"+ {failed[0].artist_name} - {failed[0].album_title} --- Reason: {failed[1]}")
                print("")
                logger.info(f":: Failed downloads exported to: {startup.get_appdata_dir()}/failed.csv")
            else:
                logger.info("Downloads complete!")
            if plex and (config.plex_library() != ""):
                refresh_plex(plex)
        return True

    def download(self, artist, artist_id, album_id, url, input_file, auto=True, from_date: str = None):

        def filter_artist_by_record_type(artist):
            album_api = self.dz.api.get_artist_albums(artist['id'])['data']
            filtered_albums = []
            for album in album_api:
                if (album['record_type'] == config.record_type()) or config.record_type() == "all":
                    album_date = dates.format_date_string(album['release_date'])
                    if self.from_release_date:
                        if album_date >= self.from_release_date:
                            filtered_albums.append(album)
                    else:
                        filtered_albums.append(album)
            return filtered_albums

        def get_api_result(artist=None, artist_id=None, album_id=None):
            if artist:
                try:
                    return self.dz.api.search_artist(artist, limit=1)["data"][0]  # TODO handle incorrect artist
                except (deezer.api.DataException, IndexError):
                    logger.error(f"Artist {artist} not found.")
            if artist_id:
                try:
                    return self.dz.api.get_artist(artist_id)
                except (deezer.api.DataException, IndexError):
                    logger.error(f"Artist ID {artist_id} not found.")
            if album_id:
                try:
                    return self.dz.api.get_album(album_id)
                except (deezer.api.DataException, IndexError):
                    logger.error(f"Album ID {album_id} not found.")

        def queue_filtered_releases(api_object):
            filtered = filter_artist_by_record_type(api_object)
            for album in filtered:
                if not queue_item_exists(album['id']):
                    self.queue_list.append(QueueItem(artist=api_object, album=album))

        def queue_item_exists(i):
            for q in self.queue_list:
                if q.album_id == i:
                    logger.debug(f"Album ID {i} is already in queue")
                    self.duplicate_id_count += 1
                    return True
            return False

        def process_artist_by_name(name):
            artist_result = get_api_result(artist=name)
            if not artist_result:
                return
            logger.debug(f"Requested Artist: '{name}', Found: '{artist_result['name']}'")
            if artist_result:
                queue_filtered_releases(artist_result)

        def process_artist_by_id(i):
            artist_id_result = get_api_result(artist_id=i)
            if not artist_id_result:
                return
            logger.debug(f"Requested Artist ID: {i}, Found: {artist_id_result['name']}")
            if artist_id_result:
                queue_filtered_releases(artist_id_result)

        def process_album_by_id(i):
            logger.debug("Processing album by name")
            album_id_result = get_api_result(album_id=i)
            if not album_id_result:
                return
            logger.debug(f"Requested Album: {i}, "
                         f"Found: {album_id_result['artist']['name']} - {album_id_result['title']}")
            if album_id_result and not queue_item_exists(album_id_result['id']):
                self.queue_list.append(QueueItem(album=album_id_result))

        def process_playlist_by_id(id):
            playlist_api = self.dz.api.get_playlist(id)
            self.queue_list.append(QueueItem(playlist=playlist_api))

        def extract_id_from_url(url):
            id_group = ['artist', 'album', 'playlist']
            for group in id_group:
                id_type = group
                try:
                    id = int(url.split(f'/{group}/')[1])
                    logger.debug(f"Extracted group={id_type}, id={id}")
                    return id_type, id
                except (IndexError, ValueError):
                    continue
            return False, False

        logger.info("Queueing releases, this might take awhile...")

        if from_date:
            logger.debug(f"Getting releases that were released on or after {from_date}")
            from_date_datetime = validate.validate_date(from_date)
            if from_date_datetime:
                self.from_release_date = from_date_datetime
            else:
                return logger.error(f"The date you entered is invalid: {from_date}")

        if artist:
            [process_artist_by_name(a) for a in artist]

        if artist_id:
            [process_artist_by_id(i) for i in artist_id]

        if album_id:
            [process_album_by_id(i) for i in album_id]

        if input_file:
            logger.info(f"Reading from file {input_file}")
            if Path(input_file).exists():
                artist_list = utils.dataprocessor.read_file_as_csv(input_file)
                artist_int_list, artist_str_list = utils.dataprocessor.process_input_file(artist_list)
                if artist_str_list:
                    for artist in artist_str_list:
                        process_artist_by_name(artist)
                if artist_int_list:
                    for artist in artist_int_list:
                        process_artist_by_id(artist)

        if url:
            logger.debug("Processing URLs")
            for u in url:
                egroup, eid = extract_id_from_url(u)
                if not egroup or not eid:
                    logger.error(f"Invalid URL -- {u}")
                    continue

                if egroup == "artist":
                    process_artist_by_id(eid)
                elif egroup == "album":
                    process_album_by_id(eid)
                elif egroup == "playlist":
                    process_playlist_by_id(eid)

        if self.duplicate_id_count > 0:
            logger.info(f"Cleaned up {self.duplicate_id_count} duplicate release(s). See log for additional info.")

        if auto:
            if len(self.queue_list):
                self.download_queue()
            else:
                logger.info("No releases found matching applied filters.")



import logging
import os
import sys
from pathlib import Path

import requests
from packaging.version import parse as parse_version

logger = logging.getLogger(__name__)


def get_appdata_root():

    home_dir = Path.home()

    if os.getenv("XDG_CONFIG_HOME"):
        appdata_dir = Path(os.getenv("XDG_CONFIG_HOME"))
    elif os.getenv("APPDATA"):
        appdata_dir = Path(os.getenv("APPDATA"))
    elif sys.platform.startswith('darwin'):
        appdata_dir = home_dir / 'Library' / 'Application Support'
    else:
        appdata_dir = home_dir / '.config'

    return appdata_dir


def get_appdata_dir():
    """
    Get appdata directory where configuration and data is stored
    """
    return get_appdata_root() / 'deemon'


def get_backup_dir():
    return Path(get_appdata_dir() / "backups")


def init_appdata_dir(appdata):
    Path(appdata / 'logs').mkdir(parents=True, exist_ok=True)
    Path(appdata / 'backups').mkdir(exist_ok=True)


def get_config():
    return get_appdata_dir() / 'config.json'


def get_database():
    return get_appdata_dir() / 'deemon.db'


def get_log_file():
    """
    Get path to log file
    """
    return Path(get_appdata_dir() / 'logs' / 'deemon.log')


def get_latest_version(release_type):
    latest_ver = "https://pypi.org/pypi/deemon/json"

    try:
        response = requests.get(latest_ver)
    except requests.exceptions.ConnectionError:
        return

    latest_stable = parse_version(response.json()['info']['version'])

    if release_type == "beta":
        all_releases = [parse_version(x) for x in response.json()['releases']]
        sorted_releases = sorted(all_releases, reverse=True)
        for release in sorted_releases:
            if "b" in str(release) or "rc" in str(release):
                if release > latest_stable:
                    return release
                else:
                    return latest_stable
    else:
        return latest_stable

def get_changelog(ver: str):
    try:
        response = requests.get("https://api.github.com/repos/digitalec/"
                                "deemon/releases")
    except requests.exceptions.ConnectionError:
        return print("Unable to reach GitHub API")
    
    for release in response.json():
        if release['name'] == ver:
            return print(release['body'])
    return print(f"Changelog for v{ver} was not found.")


import logging
from csv import reader

logger = logging.getLogger(__name__)


def read_file_as_csv(file):
    with open(file, 'r', encoding="utf-8-sig", errors="replace") as f:
        make_csv = f.read()
        csv_to_list = make_csv.split('\n')
        sorted_list = sorted(list(filter(None, csv_to_list)))
        return sorted_list


def process_input_file(artist_list):
    logger.debug("Processing file contents")
    try:
        artists = [int(x) for x in artist_list]
        logger.debug(f"File detected as containing {len(artists)} artist IDs")
    except ValueError:
        artists = [x for x in artist_list]
        logger.debug(f"File detected as containing {len(artists)} artist names")
    return artists


def csv_to_list(all_artists) -> list:
    """
    Separate artists and replace delimiter to find artists containing commas in their name
    """
    all_artists = [x for x in all_artists]
    processed_artists = []
    for artist in all_artists:
        if artist[-1] == ',':
            processed_artists.append(artist[:-1] + "|")
        else:
            processed_artists.append(artist)
    processed_artists = ' '.join(processed_artists)
    processed_artists = processed_artists.split('|')

    result = []
    csv_artists = reader(processed_artists, delimiter="|")
    for line in csv_artists:
        combined_line = ([x.lstrip() for x in line])
        result.append(','.join(combined_line))
    return (result)



import json
import logging
from datetime import datetime

import deezer.errors
from deezer import Deezer

from deemon.core.config import Config as config

logger = logging.getLogger(__name__)


class PlatformAPI:

    def __init__(self, platform: str = "deezer-api"):
        self.max_threads = 2
        self.platform = platform
        self.api = self.set_platform()

    def debugger(self, message: str, payload = None):
        if config.debug_mode():
            if not payload:
                payload = ""
            logger.debug(f"DEBUG_MODE: {message} {str(payload)}")

    def set_platform(self):
        if self.platform == "deezer-gw":
            dz = Deezer()
            if dz.login_via_arl(config.arl()):
                self.max_threads = 50
                logger.debug(f"Login OK, max_threads set to {self.max_threads}")
                api_obj = dz.gw
            else:
                logger.warning("[!] Falling back to standard API (expired ARL?)")
                self.platform = "deezer-api"
                api_obj = dz.api
        else:
            dz = Deezer()
            api_obj = dz.api

        logger.debug(f"API in use: {self.platform}, thread count set to: {self.max_threads}")
        return api_obj

    def search_artist(self, query: str, limit: int = 1) -> dict:
        """
        Return a list of dictionaries from API containing {'id': int, 'name': str}
        """
        if self.platform == "deezer-gw":
            result = self.api.search(query=query, limit=limit)['ARTIST']['data']
            api_result = []
            for r in result:
                api_result.append({'id': int(r['ART_ID']), 'name': r['ART_NAME']})
        else:
            api_result = self.api.search_artist(query=query, limit=limit)['data']

        return {'query': query, 'results': api_result}

    def get_artist_by_id(self, query: int, limit: int = 1) -> dict:
        """
        Return a dictionary from API containing {'id': int, 'name': str}
        """
        if self.platform == "deezer-gw":
            try:
                result = self.api.get_artist(query)
            except deezer.errors.GWAPIError as e:
                logger.debug(f"API error: {e}")
                return {}
            return {'id': int(result['ART_ID']), 'name': result['ART_NAME']}
        else:
            try:
                result = self.api.get_artist(query)
            except deezer.errors.DataException as e:
                logger.debug(f"API error: {e}")
                return {}
            return {'id': result['id'], 'name': result['name']}

    def get_artist_albums(self, query: dict, limit: int = -1):
        """
        Return a list of dictionaries from API containing
        """
        self.debugger("RefreshArtist", query['artist_name'])
        if self.platform == "deezer-gw":
            try:
                result = self.api.get_artist_discography(art_id=query['artist_id'], limit=limit)['data']
            except deezer.errors.GWAPIError as e:
                if "UNKNOWN" in str(e):
                    logger.error(f"   [!] Artist discography is not available for "
                                 f"{query['artist_name']} ({query['artist_id']})")
                else:
                    logger.error(f"An error occured while attempting to get the discography for "
                                 f"{query['artist_id']} ({query['artist_id']})")
                query['releases'] = []
                return query
            api_result = []
            for r in result:
                if r['ART_ID'] == str(query['artist_id']) and r['ARTISTS_ALBUMS_IS_OFFICIAL']:
                    # TYPE 0 - single, TYPE 1 - album, TYPE 2 - compilation, TYPE 3 - ep
                    if r['TYPE'] == '0':
                        r['TYPE'] = "single"
                    elif r['TYPE'] == '3':
                        r['TYPE'] = "ep"
                    else:
                        r['TYPE'] = "album"

                    if r['DIGITAL_RELEASE_DATE'] != "0000-00-00":
                        release_date = r['DIGITAL_RELEASE_DATE']
                    elif r['ORIGINAL_RELEASE_DATE'] != "0000-00-00":
                        release_date = r['ORIGINAL_RELEASE_DATE']
                    elif r['PHYSICAL_RELEASE_DATE'] != "0000-00-00":
                        release_date = r['PHYSICAL_RELEASE_DATE']
                    else:
                        # In the event of an unknown release date, set it to today's date
                        # See album ID: 417403
                        logger.warning(f"   [!] Found release without release date, assuming today: "
                                       f"{query['artist_name']} - {r['ALB_TITLE']}")
                        release_date = datetime.strftime(datetime.today(), "%Y-%m-%d")
                    
                    if config.extra_release_info():
                        album_details = self.api.get_album(r['ALB_ID'])
                        if album_details.get('LABEL_NAME'):
                            r['LABEL_NAME'] = album_details['LABEL_NAME']
                    
                    if not r.get('LABEL_NAME'):
                        r['LABEL_NAME'] = None

                    api_result.append(
                        {
                            'id': int(r['ALB_ID']),
                            'title': r['ALB_TITLE'],
                            'release_date': release_date,
                            'explicit_lyrics': r['EXPLICIT_ALBUM_CONTENT']['EXPLICIT_LYRICS_STATUS'],
                            'record_type': r['TYPE'],
                            'label': r['LABEL_NAME']
                         }
                    )
        else:
            api_result = self.api.get_artist_albums(artist_id=query['artist_id'], limit=limit)['data']

            for release in api_result:
                if config.extra_release_info():
                    album_details = self.api.get_album(release['id'])
                    if album_details.get('label'):
                        release['label'] = album_details['label']
                if not release.get('label'):
                    release['label'] = None

        query['releases'] = api_result
        return query

    def get_playlist(self, query: int):
        api_result = Deezer().api.get_playlist(query)
        return {'id': query, 'title': api_result['title'],
                'link': f"https://deezer.com/playlist/{str(api_result['id'])}"}

    def get_playlist_tracks(self, query: dict):
        track_list = []
        api_result = Deezer().api.get_playlist(query['id'])
        for track in api_result['tracks']['data']:
            track_list.append({'id': track['id'], 'title': track['title'],
                               'artist_id': track['artist']['id'],
                               'artist_name': track['artist']['name']})
        query['tracks'] = track_list
        return query



import logging
import sqlite3
import time
from datetime import datetime
from pathlib import Path

from packaging.version import parse as parse_version

from deemon import __dbversion__
from deemon.core.config import Config as config
from deemon.utils import startup, performance

logger = logging.getLogger(__name__)


class Database(object):

    def __init__(self):
        self.conn = None
        self.cursor = None
        self.db = startup.get_database()

        if not Path(self.db).exists():
            self.connect()
            self.create_new_database()
        else:
            self.connect()

    def __enter__(self):
        return self

    def __exit__(self):
        self.close()

    @staticmethod
    def dict_factory(cursor, row):
        d = {}
        for idx, col in enumerate(cursor.description):
            d[col[0]] = row[idx]
        return d

    def connect(self):
        try:
            self.conn = sqlite3.connect(self.db)
            self.conn.row_factory = self.dict_factory
            self.cursor = self.conn.cursor()
        except sqlite3.OperationalError as e:
            logger.error(f"Error opening database: {e}")

    def close(self):
        if self.conn:
            self.conn.commit()
            self.conn.close()

    def commit(self):
        self.conn.commit()

    def commit_and_close(self):
        self.commit()
        self.close()

    def create_new_database(self):
        logger.debug("DATABASE CREATION IN PROGRESS!")

        self.query("CREATE TABLE monitor ("
                   "'artist_id' INTEGER,"
                   "'artist_name' TEXT,"
                   "'bitrate' TEXT,"
                   "'record_type' TEXT,"
                   "'alerts' INTEGER,"
                   "'profile_id' INTEGER DEFAULT 1,"
                   "'download_path' TEXT,"
                   "'refreshed' INTEGER DEFAULT 0,"
                   "'trans_id' INTEGER)")

        self.query("CREATE TABLE playlists ("
                   "'id' INTEGER UNIQUE,"
                   "'title' TEXT,"
                   "'url' TEXT,"
                   "'bitrate' TEXT,"
                   "'alerts' INTEGER,"
                   "'profile_id' INTEGER DEFAULT 1,"
                   "'download_path' TEXT,"
                   "'refreshed' INTEGER DEFAULT 0,"
                   "'trans_id' INTEGER)")

        self.query("CREATE TABLE playlist_tracks ("
                   "'track_id' INTEGER,"
                   "'playlist_id' INTEGER,"
                   "'artist_id' INTEGER,"
                   "'artist_name' TEXT,"
                   "'track_name' TEXT,"
                   "'profile_id' INTEGER DEFAULT 1,"
                   "'track_added' TEXT,"
                   "'trans_id' INTEGER)")

        self.query("CREATE TABLE releases ("
                   "'artist_id' INTEGER,"
                   "'artist_name' TEXT,"
                   "'album_id' INTEGER,"
                   "'album_name' TEXT,"
                   "'album_release' TEXT,"
                   "'album_added' INTEGER,"
                   "'explicit' INTEGER,"
                   "'label' TEXT,"
                   "'profile_id' INTEGER DEFAULT 1,"
                   "'future_release' INTEGER DEFAULT 0,"
                   "'trans_id' INTEGER)")

        self.query("CREATE TABLE 'deemon' ("
                   "'property' TEXT,"
                   "'value' TEXT)")

        self.query("CREATE TABLE 'profiles' ("
                   "'id' INTEGER,"
                   "'name' TEXT,"
                   "'email' TEXT,"
                   "'alerts' INTEGER,"
                   "'bitrate' TEXT,"
                   "'record_type' TEXT,"
                   "'plex_baseurl' TEXT,"
                   "'plex_token' TEXT,"
                   "'plex_library' TEXT,"
                   "'download_path' TEXT,"
                   "PRIMARY KEY('id' AUTOINCREMENT))")

        self.query("CREATE TABLE transactions ("
                   "'id' INTEGER,"
                   "'timestamp' INTEGER,"
                   "'profile_id' INTEGER DEFAULT 1,"
                   "PRIMARY KEY('id' AUTOINCREMENT))")

        self.query("CREATE UNIQUE INDEX 'idx_property' ON 'deemon' ('property')")
        self.query(f"INSERT INTO 'deemon' ('property', 'value') VALUES ('version', '{__dbversion__}')")
        self.query("INSERT OR REPLACE INTO 'deemon' ('property', 'value') VALUES ('latest_ver', '')")
        self.query("INSERT INTO 'deemon' ('property', 'value') VALUES ('last_update_check', 0)")
        self.query("INSERT INTO 'deemon' ('property', 'value') VALUES ('release_channel', 'stable')")
        self.query("INSERT INTO 'profiles' ('name') VALUES ('default')")
        self.commit()

    def get_latest_ver(self):
        return self.query("SELECT value FROM deemon WHERE property = 'latest_ver'").fetchone()['value']

    def get_db_version(self):
        try:
            version = self.query(f"SELECT value FROM deemon WHERE property = 'version'").fetchone()['value']
        except sqlite3.OperationalError:
            version = '0.0.0'

        logger.debug(f"Database version {version}")
        return version

    def do_upgrade(self):
        current_ver = parse_version(self.get_db_version())
        app_db_version = parse_version(__dbversion__)

        if current_ver == app_db_version:
            return

        logger.debug("DATABASE UPGRADE IN PROGRESS!")
        # Upgrade database v1.0 to v1.1
        if current_ver < parse_version("1.1"):
            sql_playlists = ("CREATE TABLE IF NOT EXISTS 'playlists' "
                             "('id' INTEGER UNIQUE, 'title' TEXT, 'url' TEXT)")

            sql_playlist_tracks = ("CREATE TABLE IF NOT EXISTS 'playlist_tracks' "
                                   "('track_id' INTEGER, 'playlist_id' INTEGER, 'artist_id' INTEGER, "
                                   "'artist_name' TEXT, 'track_name' TEXT, 'track_added' TEXT)")
            self.query(sql_playlists)
            self.query(sql_playlist_tracks)
            self.query("INSERT OR REPLACE INTO 'deemon' ('property', 'value') VALUES ('version', '1.1')")
            self.commit()

        # Upgrade database v1.1 to v1.3
        if current_ver < parse_version("1.3"):
            sql_playlists_1 = "ALTER TABLE playlists ADD COLUMN bitrate TEXT"
            sql_playlists_2 = "ALTER TABLE playlists ADD COLUMN alerts INTEGER"
            sql_updatever = "INSERT OR REPLACE INTO 'deemon' ('property', 'value') VALUES ('version', '1.3')"
            self.query(sql_playlists_1)
            self.query(sql_playlists_2)
            self.query(sql_updatever)
            self.commit()

        if current_ver < parse_version("2.1"):
            self.query("INSERT OR REPLACE INTO 'deemon' ('property', 'value') VALUES ('last_update_check', 0)")
            self.query("INSERT OR REPLACE INTO 'deemon' ('property', 'value') VALUES ('version', '2.1')")
            self.commit()

        if current_ver < parse_version("2.2"):
            self.query("ALTER TABLE monitor ADD COLUMN download_path TEXT")
            self.query("ALTER TABLE playlists ADD COLUMN download_path TEXT")
            self.query("INSERT OR REPLACE INTO 'deemon' ('property', 'value') VALUES ('version', '2.2')")
            self.commit()

        # Upgrade database to v3
        if current_ver < parse_version("3.0"):
            self.query("CREATE TABLE 'profiles' ("
                       "'id' INTEGER,"
                       "'name' TEXT,"
                       "'email' TEXT,"
                       "'alerts' INTEGER,"
                       "'bitrate' TEXT,"
                       "'record_type' TEXT,"
                       "'plex_baseurl' TEXT,"
                       "'plex_token' TEXT,"
                       "'plex_library' TEXT,"
                       "'download_path' TEXT,"
                       "PRIMARY KEY('id' AUTOINCREMENT))")
            self.query("INSERT INTO 'profiles' ('name') VALUES ('default')")
            self.query("ALTER TABLE monitor ADD COLUMN profile_id INTEGER DEFAULT 1")
            self.query("ALTER TABLE releases ADD COLUMN profile_id INTEGER DEFAULT 1")
            self.query("ALTER TABLE playlists ADD COLUMN profile_id INTEGER DEFAULT 1")
            self.query("ALTER TABLE playlist_tracks ADD COLUMN profile_id INTEGER DEFAULT 1")
            self.query("CREATE TABLE monitor_tmp ("
                       "'artist_id' INTEGER,"
                       "'artist_name' TEXT,"
                       "'bitrate' TEXT,"
                       "'record_type' TEXT,"
                       "'alerts' INTEGER,"
                       "'download_path' TEXT,"
                       "'profile_id' INTEGER DEFAULT 1)")
            self.query("INSERT INTO monitor_tmp SELECT * FROM monitor")
            self.query("DROP TABLE monitor")
            self.query("ALTER TABLE monitor_tmp RENAME TO monitor")
            self.query("INSERT OR REPLACE INTO 'deemon' ('property', 'value') VALUES ('latest_ver', '')")
            self.query("INSERT OR REPLACE INTO 'deemon' ('property', 'value') VALUES ('version', '3.0')")
            self.commit()

        if current_ver < parse_version("3.1"):
            self.query("CREATE TABLE transactions ("
                       "'id' INTEGER,"
                       "'timestamp' INTEGER,"
                       "'profile_id' INTEGER DEFAULT 1,"
                       "PRIMARY KEY('id' AUTOINCREMENT))")
            self.query("ALTER TABLE monitor ADD COLUMN trans_id INTEGER")
            self.query("ALTER TABLE monitor ADD COLUMN refreshed INTEGER DEFAULT 0")
            self.query("ALTER TABLE playlists ADD COLUMN trans_id INTEGER")
            self.query("ALTER TABLE playlists ADD COLUMN refreshed INTEGER DEFAULT 0")
            self.query("ALTER TABLE releases ADD COLUMN trans_id")
            self.query("ALTER TABLE playlist_tracks ADD COLUMN trans_id")
            self.query("UPDATE monitor SET bitrate = '128' WHERE bitrate = '1'")
            self.query("UPDATE monitor SET bitrate = '320' WHERE bitrate = '3'")
            self.query("UPDATE monitor SET bitrate = 'FLAC' WHERE bitrate = '9'")
            self.query("UPDATE playlists SET bitrate = '128' WHERE bitrate = '1'")
            self.query("UPDATE playlists SET bitrate = '320' WHERE bitrate = '3'")
            self.query("UPDATE playlists SET bitrate = 'FLAC' WHERE bitrate = '9'")
            self.query("INSERT OR REPLACE INTO 'deemon' ('property', 'value') VALUES ('version', '3.1')")
            self.query("INSERT INTO 'deemon' ('property', 'value') VALUES ('release_channel', 'stable')")
            self.commit()

        if current_ver < parse_version("3.2"):
            self.query("CREATE TABLE playlists_tmp ("
                       "'id' INTEGER UNIQUE,"
                       "'title' TEXT,"
                       "'url' TEXT,"
                       "'bitrate' TEXT,"
                       "'alerts' INTEGER,"
                       "'profile_id' INTEGER DEFAULT 1,"
                       "'download_path' TEXT,"
                       "'refreshed' INTEGER DEFAULT 0,"
                       "'trans_id' INTEGER)")
            self.query("INSERT INTO playlists_tmp SELECT * FROM playlists")
            self.query("DROP TABLE playlists")
            self.query("ALTER TABLE playlists_tmp RENAME TO playlists")
            self.query("INSERT OR REPLACE INTO 'deemon' ('property', 'value') VALUES ('version', '3.2')")
            self.commit()
        
        if current_ver < parse_version("3.3"):
            self.query("ALTER TABLE releases ADD COLUMN label TEXT")
            self.query("INSERT OR REPLACE INTO 'deemon' ('property', 'value') VALUES ('version', '3.3')")
            self.commit()
            logger.debug(f"Database upgraded to version 3.3")

    def query(self, query, values=None):
        if values is None:
            values = {}
        return self.cursor.execute(query, values)

    def reset_future(self, album_id):
        logger.debug("Clearing future_release flag from " + str(album_id))
        values = {'album_id': album_id}
        sql = "UPDATE 'releases' SET future_release = 0 WHERE album_id = :album_id"
        self.query(sql, values)

    def get_all_monitored_artists(self):
        vals = {'profile_id': config.profile_id()}
        return self.query(f"SELECT * FROM monitor WHERE profile_id = :profile_id "
                          f"ORDER BY artist_name COLLATE NOCASE ASC", vals).fetchall()

    def get_monitored_artist_by_id(self, artist_id: int):
        values = {'id': artist_id, 'profile_id': config.profile_id()}
        return self.query(f"SELECT * FROM monitor WHERE artist_id = :id AND profile_id = :profile_id",
                          values).fetchone()

    def get_monitored_artist_by_name(self, name: str):
        values = {'name': name, 'profile_id': config.profile_id()}
        return self.query(f"SELECT * FROM monitor WHERE artist_name = :name COLLATE NOCASE "
                          f"AND profile_id = :profile_id", values).fetchone()

    def get_all_monitored_playlist_ids(self):
        vals = {'profile_id': config.profile_id()}
        query = self.query("SELECT id FROM playlists WHERE profile_id = :profile_id", vals).fetchall()
        return [v for x in query for v in x.values()]

    def get_all_monitored_playlists(self):
        vals = {'profile_id': config.profile_id()}
        return self.query("SELECT * FROM playlists WHERE profile_id = :profile_id "
                          "ORDER BY title COLLATE NOCASE ASC", vals).fetchall()

    def get_monitored_playlist_by_id(self, playlist_id):
        values = {'id': playlist_id, 'profile_id': config.profile_id()}
        return self.query("SELECT * FROM playlists WHERE id = :id AND profile_id = :profile_id", values).fetchone()

    def get_monitored_playlist_by_name(self, title):
        values = {'title': title, 'profile_id': config.profile_id()}
        return self.query("SELECT * FROM playlists WHERE title = :title COLLATE NOCASE "
                          "AND profile_id = :profile_id", values).fetchone()

    def monitor_artist(self, artist: dict, artist_config: dict):
        self.new_transaction()
        vals = {
            'artist_id': artist['id'], 'artist_name': artist['name'], 'bitrate': artist_config['bitrate'],
            'record_type': artist_config['record_type'], 'alerts': artist_config['alerts'],
            'download_path': artist_config['download_path'], 'profile_id': config.profile_id(),
            'trans_id': config.transaction_id()
        }
        query = ("INSERT INTO monitor "
                 "(artist_id, artist_name, bitrate, record_type, alerts, download_path, profile_id, trans_id) "
                 "VALUES "
                 "(:artist_id, :artist_name, :bitrate, :record_type, :alerts, :download_path, :profile_id, :trans_id)")
        self.query(query, vals)
        self.commit()

    def get_artist_releases(self, artist_id=None):
        sql_values = {'artist_id': artist_id, 'profile_id': config.profile_id()}
        if artist_id:
            query = "SELECT album_id FROM 'releases' WHERE artist_id = :artist_id AND profile_id = :profile_id"
        else:
            query = "SELECT album_id FROM 'releases' WHERE profile_id = :profile_id"
        return self.query(query, sql_values).fetchall()

    def get_future_releases(self):
        vals = {'profile_id': config.profile_id()}
        return self.query("SELECT * FROM releases "
                          "WHERE future_release = 1 AND profile_id = :profile_id", vals).fetchall()

    def get_playlist_tracks(self, playlist_id):
        sql_values = {'playlist_id': playlist_id, 'profile_id': config.profile_id()}
        query = "SELECT * FROM 'playlist_tracks' WHERE playlist_id = :playlist_id AND profile_id = :profile_id"
        return self.query(query, sql_values).fetchall()

    def get_track_from_playlist(self, playlist_id, track_id):
        values = {'pid': playlist_id, 'tid': track_id, 'profile_id': config.profile_id()}
        query = "SELECT * FROM 'playlist_tracks' WHERE track_id = :tid AND playlist_id = :pid AND profile_id = :profile_id"
        result = self.query(query, values).fetchone()
        if result:
            return True

    def monitor_playlist(self, api_result):
        self.new_transaction()
        values = {'id': api_result['id'], 'title': api_result['title'], 'url': api_result['link'],
                  'bitrate': api_result['bitrate'], 'alerts': api_result['alerts'],
                  'download_path': api_result['download_path'], 'profile_id': config.profile_id(),
                  'trans_id': config.transaction_id()}
        query = ("INSERT INTO playlists ('id', 'title', 'url', 'bitrate', 'alerts', 'download_path',"
                 "'profile_id', 'trans_id') "
                 "VALUES (:id, :title, :url, :bitrate, :alerts, :download_path, :profile_id, :trans_id)")
        self.query(query, values)
        self.commit()

    def remove_monitored_artist(self, id: int = None):
        values = {'id': id, 'profile_id': config.profile_id()}
        self.query("DELETE FROM monitor WHERE artist_id = :id AND profile_id = :profile_id", values)
        self.query("DELETE FROM releases WHERE artist_id = :id AND profile_id = :profile_id", values)
        self.commit()

    def remove_monitored_playlists(self, id: int = None):
        values = {'id': id, 'profile_id': config.profile_id()}
        self.query("DELETE FROM playlists WHERE id = :id AND profile_id = :profile_id", values)
        self.query("DELETE FROM playlist_tracks WHERE playlist_id = :id AND profile_id = :profile_id", values)
        self.commit()

    def get_specified_artist(self, artist):
        values = {'artist': artist, 'profile_id': config.profile_id()}
        if type(artist) is int:
            return self.query("SELECT * FROM monitor WHERE artist_id = :artist "
                              "AND profile_id = :profile_id", values).fetchone()
        else:
            return self.query("SELECT * FROM monitor WHERE artist_name = ':artist' "
                              "AND profile_id = :profile_id COLLATE NOCASE", values).fetchone()

    def set_all_artists_refreshed(self):
        self.query("UPDATE monitor SET refreshed = 1 WHERE refreshed = 0")
        
    def set_all_playlists_refreshed(self):
        self.query("UPDATE playlists SET refreshed = 1 WHERE refreshed = 0")

    def add_new_releases(self, values):
        self.new_transaction()
        sql = (f"INSERT INTO releases ('artist_id', 'artist_name', 'album_id', 'album_name', 'album_release', "
               f"'album_added', 'future_release', 'explicit', 'label', 'profile_id', 'trans_id') "
               f"VALUES (:artist_id, :artist_name, :id, :title, :release_date, {int(time.time())}, :future, "
               f":explicit_lyrics, :label, {config.profile_id()}, {config.transaction_id()})")
        self.cursor.executemany(sql, values)
        self.set_all_artists_refreshed()

    def add_new_playlist_releases(self, values):
        self.new_transaction()
        sql = (f"INSERT INTO playlist_tracks ('artist_id', 'artist_name', 'track_id', 'track_name', 'playlist_id', "
               f"'track_added', 'profile_id', 'trans_id') VALUES (:artist_id, :artist_name, :id, :title, :playlist_id, "
               f"{int(time.time())}, {config.profile_id()}, {config.transaction_id()})")
        self.cursor.executemany(sql, values)
        self.set_all_playlists_refreshed()

    def show_new_releases(self, from_date_ts, now_ts):
        today_date = datetime.utcfromtimestamp(now_ts).strftime('%Y-%m-%d')
        from_date = datetime.utcfromtimestamp(from_date_ts).strftime('%Y-%m-%d')
        values = {'from': from_date, 'today': today_date, 'profile_id': config.profile_id()}
        sql = "SELECT * FROM 'releases' WHERE album_release >= :from AND album_release <= :today AND profile_id = :profile_id"
        return self.query(sql, values).fetchall()

    def get_album_by_id(self, album_id):
        values = {'id': album_id, 'profile_id': config.profile_id()}
        sql = "SELECT * FROM 'releases' WHERE album_id = :id AND profile_id = :profile_id"
        return self.query(sql, values).fetchone()

    def reset_database(self):
        self.query("DELETE FROM monitor")
        self.query("DELETE FROM releases")
        self.query("DELETE FROM playlists")
        self.query("DELETE FROM playlist_tracks")
        self.query("DELETE FROM transactions")
        self.commit()
        logger.info("Database has been reset")

    def update_artist(self, settings: dict):
        self.query("UPDATE monitor SET bitrate = :bitrate, alerts = :alerts, record_type = :record_type,"
                   "download_path = :download_path WHERE artist_id = :artist_id AND profile_id = :profile_id", settings)
        self.commit()

    def add_playlist_track(self, playlist, track):
        self.new_transaction()
        values = {'pid': playlist['id'], 'tid': track['id'], 'tname': track['title'], 'aid': track['artist']['id'],
                  'aname': track['artist']['name'], 'time': int(time.time()), 'profile_id': config.profile_id(),
                  'trans_id': config.transaction_id()}
        query = ("INSERT INTO 'playlist_tracks' "
                 "('track_id', 'playlist_id', 'artist_id', 'artist_name', 'track_name', 'track_added', 'profile_id',"
                 "'trans_id') VALUES (:tid, :pid, :aid, :aname, :tname, :time, :profile_id, :trans_id)")
        return self.query(query, values)

    def create_profile(self, settings: dict):
        self.query("INSERT INTO profiles (name, email, alerts, bitrate, record_type, plex_baseurl, plex_token,"
                   "plex_library, download_path) VALUES (:name, :email, :alerts, :bitrate, :record_type,"
                   ":plex_baseurl, :plex_token, :plex_library, :download_path)", settings)
        self.commit()

    def delete_profile(self, profile_name: str):
        profile = self.get_profile(profile_name)
        vals = {'profile_id': profile['id']}
        self.query("DELETE FROM monitor WHERE profile_id = :profile_id", vals)
        self.query("DELETE FROM releases WHERE profile_id = :profile_id", vals)
        self.query("DELETE FROM playlists WHERE profile_id = :profile_id", vals)
        self.query("DELETE FROM playlist_tracks WHERE profile_id = :profile_id", vals)
        self.query("DELETE FROM profiles WHERE id = :profile_id", vals)
        self.commit()

    def get_all_profiles(self):
        return self.query("SELECT * FROM profiles").fetchall()

    def get_profile(self, profile_name: str):
        vals = {'profile': profile_name}
        return self.query("SELECT * FROM profiles WHERE name = :profile COLLATE NOCASE", vals).fetchone()

    def get_profile_by_id(self, profile_id: int):
        vals = {'profile_id': profile_id}
        return self.query("SELECT * FROM profiles WHERE id = :profile_id", vals).fetchone()

    def update_profile(self, settings: dict):
        self.query("UPDATE profiles SET name = :name, email = :email, alerts = :alerts, bitrate = :bitrate,"
                   "record_type = :record_type, plex_baseurl = :plex_baseurl, plex_token = :plex_token,"
                   "plex_library = :plex_library, download_path = :download_path "
                   "WHERE id = :id", settings)
        self.commit()

    def last_update_check(self):
        return self.query("SELECT value FROM 'deemon' WHERE property = 'last_update_check'").fetchone()['value']

    def set_last_update_check(self):
        now = int(time.time())
        self.query(f"UPDATE deemon SET value = {now} WHERE property = 'last_update_check'")
        self.commit()

    def get_next_transaction_id(self):
        tid = self.query(f"SELECT seq FROM sqlite_sequence WHERE name = 'transactions'").fetchone()
        if not tid:
            return 0
        return tid['seq'] + 1

    def new_transaction(self):
        check_exists = self.query(f"SELECT * FROM transactions WHERE id = {config.transaction_id()}").fetchone()
        if not check_exists:
            current_time = int(time.time())
            vals = {'timestamp': current_time, 'profile_id': config.profile_id()}
            self.query(f"INSERT INTO transactions ('timestamp', 'profile_id') "
                       f"VALUES (:timestamp, :profile_id)", vals)
            self.commit()

    def rollback_last_refresh(self, rollback: int):
        vals = {'rollback': rollback, 'profile_id': config.profile_id()}
        transactions = self.query("SELECT id FROM transactions WHERE profile_id = :profile_id "
                                  f"ORDER BY id DESC LIMIT {rollback}", vals).fetchall()
        for t in transactions:
            vals = {'id': t['id'], 'profile_id': config.profile_id()}
            self.query(f"DELETE FROM monitor WHERE trans_id = :id AND profile_id = :profile_id", vals)
            self.query(f"DELETE FROM releases WHERE trans_id = :id AND profile_id = :profile_id", vals)
            self.query(f"DELETE FROM playlist_tracks WHERE trans_id = :id AND profile_id = :profile_id", vals)
            self.query(f"DELETE FROM transactions WHERE id = :id AND profile_id = :profile_id", vals)
            self.commit()

    def rollback_refresh(self, rollback: int):
        vals = {'rollback': rollback, 'profile_id': config.profile_id()}
        self.query(f"DELETE FROM monitor WHERE trans_id = {rollback} AND profile_id = :profile_id", vals)
        self.query(f"DELETE FROM releases WHERE trans_id = {rollback} AND profile_id = :profile_id", vals)
        self.query(f"DELETE FROM playlist_tracks WHERE trans_id = {rollback} AND profile_id = :profile_id", vals)
        self.query(f"DELETE FROM transactions WHERE id = {rollback} AND profile_id = :profile_id", vals)
        self.commit()

    def set_artist_refreshed(self, id):
        vals = {'id': id, 'profile_id': config.profile_id()}
        return self.query("UPDATE monitor SET refreshed = 1 WHERE artist_id = :id AND profile_id = :profile_id", vals)

    def set_playlist_refreshed(self, id):
        vals = {'id': id, 'profile_id': config.profile_id()}
        return self.query("UPDATE playlists SET refreshed = 1 WHERE id = :id AND profile_id = :profile_id", vals)

    def set_latest_version(self, version):
        vals = {'version': version}
        self.query("INSERT OR REPLACE INTO deemon (property, value) VALUES ('latest_ver', :version)", vals)
        return self.commit()

    def get_release_channel(self):
        return self.query("SELECT value FROM deemon WHERE property = 'release_channel'").fetchone()

    def set_release_channel(self):
        self.query(f"INSERT OR REPLACE INTO deemon (property, value) "
                   f"VALUES ('release_channel', '{config.release_channel()}')")
        return self.commit()

    def get_transactions(self):
        vals = {'profile_id': config.profile_id(), 'trans_limit': config.rollback_view_limit()}
        transaction_list = self.query("SELECT id, timestamp FROM transactions WHERE profile_id = :profile_id "
                                      "ORDER BY id DESC LIMIT :trans_limit", vals).fetchall()
        results = []
        for tid in transaction_list:
            vals = {'tid': tid['id'], 'profile_id': config.profile_id()}
            transaction = {}
            transaction['id'] = tid['id']
            transaction['timestamp'] = tid['timestamp']
            transaction['releases'] = self.query("SELECT album_id "
                                                 "FROM releases "
                                                 "WHERE trans_id = :tid "
                                                 "AND profile_id = :profile_id", vals).fetchall()
            transaction['playlist_tracks'] = self.query("SELECT track_id "
                                                        "FROM playlist_tracks "
                                                        "WHERE trans_id = :tid "
                                                        "AND profile_id = :profile_id", vals).fetchall()
            transaction['playlists'] = self.query("SELECT title "
                                                  "FROM playlists "
                                                  "WHERE trans_id = :tid "
                                                  "AND profile_id = :profile_id", vals).fetchall()
            transaction['monitor'] = self.query("SELECT artist_name "
                                                "FROM monitor "
                                                "WHERE trans_id = :tid "
                                                "AND profile_id = :profile_id", vals).fetchall()
            results.append(transaction)
        return results

    def get_all_monitored_artist_ids(self):
        values = {"profile_id": config.profile_id()}
        query = self.query("SELECT artist_id FROM monitor WHERE profile_id = :profile_id", values).fetchall()
        return [v for x in query for v in x.values()]

    @performance.timeit
    def get_monitored(self):
        values = {"profile_id": config.profile_id()}
        query = self.query("SELECT artist_id, artist_name FROM monitor WHERE profile_id = :profile_id",
                           values).fetchall()
        return query

    def get_unrefreshed_artists(self):
        values = {"profile_id": config.profile_id()}
        return self.query("SELECT * FROM monitor WHERE profile_id = :profile_id AND refreshed = 0", values).fetchall()

    def get_unrefreshed_playlists(self):
        values = {"profile_id": config.profile_id()}
        return self.query("SELECT * FROM playlists WHERE profile_id = :profile_id AND refreshed = 0", values).fetchall()

    def fast_monitor(self, values):
        self.cursor.executemany(
            "INSERT OR REPLACE INTO monitor (artist_id, artist_name, bitrate, record_type, alerts, profile_id, download_path, trans_id) VALUES (:id, :name, :bitrate, :record_type, :alerts, :profile_id, :download_path, :trans_id)",
            values)

    def fast_monitor_playlist(self, values):
        self.cursor.executemany(
            "INSERT OR REPLACE INTO playlists (id, title, url, bitrate, alerts, profile_id, download_path, trans_id) VALUES (:id, :title, :link, :bitrate, :alerts, :profile_id, :download_path, :trans_id)",
            values)

    def insert_multiple(self, table, values):
        self.cursor.executemany(
            f"INSERT INTO {table} (artist_id, artist_name, album_id, album_name, album_release, album_added, profile_id, future_release, trans_id) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)",
            values)

    def remove_by_name(self, values):
        self.cursor.executemany(f"DELETE FROM monitor WHERE profile_id = {config.profile_id()} AND artist_name = ?",
                                values)
        self.cursor.executemany(f"DELETE FROM releases WHERE profile_id = {config.profile_id()} AND artist_name = ?",
                                values)
        self.commit()

    def remove_by_id(self, values):
        self.cursor.executemany(f"DELETE FROM monitor WHERE profile_id = {config.profile_id()} AND artist_id = ?",
                                values)
        self.cursor.executemany(f"DELETE FROM releases WHERE profile_id = {config.profile_id()} AND artist_id = ?",
                                values)
        self.commit()

    # @performance.timeit
    def remove_specific_releases(self, values):
        self.cursor.executemany(f"DELETE FROM releases WHERE artist_id = :id AND profile_id = {config.profile_id()}",
                                values)
        self.commit()

    # @performance.timeit
    def remove_specific_playlist_tracks(self, values):
        self.cursor.executemany(
            f"DELETE FROM playlist_tracks WHERE playlist_id = :id AND profile_id = {config.profile_id()}", values)
        self.commit()



import logging
import platform
import sys
import time
from pathlib import Path

import click
from packaging.version import parse as parse_version

from deemon import __version__
from deemon.cmd import download, rollback, backup
from deemon.cmd.artistconfig import artist_lookup
from deemon.cmd.monitor import Monitor
from deemon.cmd.profile import ProfileConfig
from deemon.cmd.refresh import Refresh
from deemon.cmd.search import Search
from deemon.cmd.show import Show
from deemon.core import notifier
from deemon.core.config import Config, LoadProfile
from deemon.core.db import Database
from deemon.core.logger import setup_logger
from deemon.utils import startup, dataprocessor, validate

logger = None
config = None
db = None

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

@click.group(context_settings=CONTEXT_SETTINGS, invoke_without_command=True,
             no_args_is_help=True)
@click.option('--whats-new', is_flag=True, help="Show release notes from this version")
@click.option('-P', '--profile', help="Specify profile to run deemon as")
@click.version_option(__version__, '-V', '--version', message='deemon %(version)s')
@click.option('-v', '--verbose', is_flag=True, help="Show debug output")
def run(whats_new, verbose, profile):
    """Monitoring and alerting tool for new music releases using the Deezer API.

    deemon is a free and open source tool. To report issues or to contribute,
    please visit https://github.com/digitalec/deemon
    """
    global logger
    global config
    global db

    setup_logger(log_level='DEBUG' if verbose else 'INFO', log_file=startup.get_log_file())
    logger = logging.getLogger(__name__)
    logger.debug(f"deemon {__version__}")
    logger.debug(f"command: \"{' '.join([x for x in sys.argv[1:]])}\"")
    logger.debug("Python " + platform.python_version())
    logger.debug(platform.platform())
    logger.debug(f"deemon appdata is located at {startup.get_appdata_dir()}")
    
    if whats_new:
        return startup.get_changelog(__version__)

    print(f"Initializing deemon {__version__}\n")

    config = Config()
    db = Database()

    db.do_upgrade()
    tid = db.get_next_transaction_id()
    config.set('tid', tid, validate=False)

    if profile:
        profile_config = db.get_profile(profile)
        if profile_config:
            LoadProfile(profile_config)
        else:
            logger.error(f"Profile {profile} does not exist.")
            sys.exit(1)
    else:
        profile_config = db.get_profile_by_id(1)
        if profile_config:
            LoadProfile(profile_config)

    last_checked: int = int(db.last_update_check())

    next_check: int = last_checked + (config.check_update() * 86400)

    if config.release_channel() != db.get_release_channel()['value']:
        # If release_channel has changed, check for latest release
        logger.debug(f"Release channel changed to '{config.release_channel()}'")
        db.set_release_channel()
        last_checked = 1

    if time.time() >= next_check or last_checked == 0:
        logger.debug(f"Checking for updates ({config.release_channel()})...")
        config.set('update_available', 0, False)
        latest_ver = str(startup.get_latest_version(config.release_channel()))
        if latest_ver:
            db.set_latest_version(latest_ver)
        db.set_last_update_check()

    new_version = db.get_latest_ver()
    if parse_version(new_version) > parse_version(__version__):
        config.set('update_available', 1, False)
        print("*" * 50)
        logger.info(f"* New version is available: v{__version__} -> v{new_version}")
        if config.release_channel() == "beta":
            logger.info("* To upgrade, run `pip install --upgrade --pre deemon`")
        else:
            logger.info("* To upgrade, run `pip install --upgrade deemon`")
        print("*" * 50)
        print("")

    config.set("start_time", int(time.time()), False)


@run.command(name='test')
def test():
    """Test email server settings by sending a test notification"""
    notification = notifier.Notify()
    notification.test()


@run.command(name='download', no_args_is_help=True)
@click.argument('artist', nargs=-1)
@click.option('-A', '--album-id', multiple=True, metavar='ID', type=int, help='Download by album ID')
@click.option('-b', '--bitrate', metavar="BITRATE", help='Set custom bitrate for this operation')
@click.option('-F', '--from-date', metavar="YYYY-MM-DD", type=str, help='Grab releases from this date forward')
@click.option('-f', '--file', metavar='FILE', help='Download batch of artists and/or artist IDs from file')
@click.option('-i', '--artist-id', multiple=True, metavar='ID', type=int, help='Download by artist ID')
@click.option('-o', '--download-path', metavar="PATH", type=str, help='Specify custom download directory')
@click.option('-t', '--record-type', metavar="TYPE", type=str, help='Specify record types to download')
@click.option('-u', '--url', metavar='URL', multiple=True, help='Download by URL of artist/album/track/playlist')
def download_command(artist, artist_id, album_id, url, file, bitrate, record_type, download_path, from_date):
    """
    Download specific artist, album ID or by URL

    \b
    Examples:
        download Mozart
        download -i 100 -t album -b 9
    """
    if bitrate:
        config.set('bitrate', bitrate)
    if download_path:
        config.set('download_path', download_path)
    if record_type:
        config.set('record_type', record_type)

    artists = dataprocessor.csv_to_list(artist) if artist else None
    artist_ids = [x for x in artist_id] if artist_id else None
    album_ids = [x for x in album_id] if album_id else None
    urls = [x for x in url] if url else None

    if download_path and download_path != "":
        if Path(download_path).exists:
            config.set('download_path', download_path)
            logger.debug(f"Download path has changed: {config.download_path()}")
        else:
            return logger.error(f"Invalid download path: {download_path}")

    dl = download.Download()
    dl.download(artists, artist_ids, album_ids, urls, file, from_date=from_date)


@run.command(name='monitor', context_settings={"ignore_unknown_options": False}, no_args_is_help=True)
@click.argument('artist', nargs=-1)
@click.option('-a', '--alerts', is_flag=True, help="Enable or disable alerts")
@click.option('-b', '--bitrate', metavar="BITRATE", help="Specify bitrate")
@click.option('-D', '--download', 'dl', is_flag=True, help='Download all releases matching record type')
@click.option('-d', '--download-path', type=str, metavar="PATH", help='Specify custom download directory')
@click.option('-I', '--import', 'im', metavar="PATH", help="Monitor artists/IDs from file or directory")
@click.option('-i', '--artist-id', is_flag=True, help="Monitor artist by ID")
@click.option('-p', '--playlist', is_flag=True, help='Monitor Deezer playlist by URL')
@click.option('-u', '--url', is_flag=True, help='Monitor artist by URL')
@click.option('-R', '--remove', is_flag=True, help='Stop monitoring an artist')
@click.option('-s', '--search', 'search_flag', is_flag=True, help='Show similar artist results to choose from')
@click.option('-T', '--time-machine', multiple=True, type=str, metavar="YYYY-MM-DD", help="Refresh newly added artists on this date")
@click.option('-t', '--record-type', metavar="TYPE", type=str, help='Specify record types to download')
def monitor_command(artist, im, playlist, bitrate, record_type, alerts, artist_id,
                    dl, remove, url, download_path, search_flag, time_machine):
    """
    Monitor artist for new releases by ID, URL or name.

    \b
    Examples:
        monitor Mozart
        monitor --artist-id 100
        monitor --url https://www.deezer.com/us/artist/000
    """
    monitor = Monitor()
    if download_path:
        if Path(download_path).exists():
            download_path = Path(download_path)
        else:
            return logger.error("Invalid download path provided")

    if time_machine:
        time_machine_dates = [x for x in time_machine]
        time_machine = []
        for d in time_machine_dates:
            validated = validate.validate_date(d)
            if not validated:
                return logger.error("Date for time machine is invalid")
            time_machine.append(validated)
        monitor.time_machine = time_machine
    
    if not alerts:
        alerts = None

    monitor.set_options(remove, dl, search_flag)
    monitor.set_config(bitrate, alerts, record_type, download_path)

    if url:
        artist_id = True
        urls = [x.replace(",", "") for x in artist]
        artist = []
        for u in urls:
            id_from_url = u.split('/artist/')
            try:
                aid = int(id_from_url[1])
            except (IndexError, ValueError):
                logger.error(f"Invalid URL -- {url}")
                return
            artist.append(aid)

    if playlist:
        urls = [x.replace(",", "") for x in artist]
        playlist_id = []
        for u in urls:
            id_from_url = u.split('/playlist/')
            try:
                aid = int(id_from_url[1])
            except (IndexError, ValueError):
                logger.error(f"Invalid playlist URL -- {url}")
                return
            playlist_id.append(aid)

    if im:
        monitor.importer(im)
    elif playlist:
        monitor.playlists(playlist_id)
    elif artist_id:
        monitor.artist_ids(dataprocessor.csv_to_list(artist))
    elif artist:
        monitor.artists(dataprocessor.csv_to_list(artist))


@run.command(name='refresh')
@click.argument('NAME', nargs=-1, type=str, required=False)
@click.option('-p', '--playlist', is_flag=True, help="Refresh a specific playlist by name")
@click.option('-s', '--skip-download', is_flag=True, help="Skips downloading of new releases")
@click.option('-T', '--time-machine', multiple=True, metavar='DATE', type=str,
              help='Refresh as if it were this date (YYYY-MM-DD)')
def refresh_command(name, playlist, skip_download, time_machine):
    """Check artists for new releases"""

    if time_machine:
        time_machine_dates = [x for x in time_machine]
        time_machine = []
        for d in time_machine_dates:
            validated = validate.validate_date(d)
            if not validated:
                return logger.error("Date for time machine is invalid")
            time_machine.append(validated)

    logger.info("Starting database refresh")
    refresh = Refresh(time_machine, skip_download)
    if playlist:
        if not len(name):
            return logger.warning("You must provide the name of a playlist")
        refresh.run(playlists=dataprocessor.csv_to_list(name))
    elif name:
        refresh.run(artists=dataprocessor.csv_to_list(name))
    else:
        refresh.run()


@click.group(name="show")
def show_command():
    """
    Show monitored artists and latest releases
    """


@show_command.command(name="artists")
@click.argument('artist', nargs=-1, required=False)
@click.option('-c', '--csv', is_flag=True, help='Output artists as CSV')
@click.option('-e', '--export', type=Path, help='Export CSV data to file; same as -ce')
@click.option('-f', '--filter', type=str, help='Specify filter for CSV output')
@click.option('-H', '--hide-header', is_flag=True, help='Hide header on CSV output')
@click.option('-i', '--artist-id', is_flag=True, help='Show artist info by artist ID')
def show_artists(artist, artist_id, csv, export, filter, hide_header):
    """Show artist info monitored by profile"""
    if artist:
        artist = ' '.join([x for x in artist])

    show = Show()
    show.monitoring(artist=True, query=artist, export_csv=csv, save_path=export, filter=filter, hide_header=hide_header,
                    is_id=artist_id)


@show_command.command(name="playlists")
@click.argument('title', nargs=-1, required=False)
@click.option('-c', '--csv', is_flag=True, help='Output artists as CSV')
@click.option('-f', '--filter', type=str, help='Specify filter for CSV output')
@click.option('-H', '--hide-header', is_flag=True, help='Hide header on CSV output')
@click.option('-i', '--playlist-id', is_flag=True, help='Show playlist info by playlist ID')
def show_artists(title, playlist_id, csv, filter, hide_header):
    """Show playlist info monitored by profile"""
    if title:
        title = ' '.join([x for x in title])

    show = Show()
    show.monitoring(artist=False, query=title, export_csv=csv, filter=filter, hide_header=hide_header, is_id=playlist_id)


@show_command.command(name="releases")
@click.argument('N', default=7)
@click.option('-f', '--future', is_flag=True, help='Display future releases')
def show_releases(n, future):
    """
    Show list of new or future releases
    """
    show = Show()
    show.releases(n, future)


run.add_command(show_command)


@run.command(name="backup")
@click.option('-i', '--include-logs', is_flag=True, help='include log files in backup')
@click.option('-r', '--restore', is_flag=True, help='Restore from existing backup')
def backup_command(restore, include_logs):
    """Backup configuration and database to a tar file"""

    if restore:
        backup.restore()
    else:
        backup.run(include_logs)


# TODO @click.option does not support nargs=-1; unable to use spaces without quotations
@run.command(name="api", help="View raw API data for artist, artist ID or playlist ID", hidden=True)
@click.option('-A', '--album-id', type=int, help='Get album ID result via API')
@click.option('-a', '--artist', type=str, help='Get artist result via API')
@click.option('-i', '--artist-id', type=int, help='Get artist ID result via API')
@click.option('-l', '--limit', type=int, help='Set max number of artist results; default=1', default=1)
@click.option('-p', '--playlist-id', type=int, help='Get playlist ID result via API')
@click.option('-r', '--raw', is_flag=True, help='Dump as raw data returned from API')
def api_test(artist, artist_id, album_id, playlist_id, limit, raw):
    """View API result - for testing purposes"""
    import deezer
    dz = deezer.Deezer()
    if artist or artist_id:
        if artist:
            result = dz.api.search_artist(artist, limit=limit)['data']
        else:
            result = dz.api.get_artist(artist_id)

        if raw:
            if isinstance(result, list):
                for row in result:
                    for key, value in row.items():
                        print(f"{key}: {value}")
                    print("\n")
            else:
                for key, value in result.items():
                    print(f"{key}: {value}")
        else:
            if isinstance(result, list):
                for row in result:
                    print(f"Artist ID: {row['id']}\nArtist Name: {row['name']}\n")
            else:
                print(f"Artist ID: {result['id']}\nArtist Name: {result['name']}")

    if album_id:
        result = dz.api.get_album(album_id)

        if raw:
            for key, value in result.items():
                print(f"{key}: {value}")
        else:
            print(f"Album ID: {result['id']}\nAlbum Title: {result['title']}")

    if playlist_id:
        result = dz.api.get_playlist(playlist_id)

        if raw:
            for key, value in result.items():
                print(f"{key}: {value}")
        else:
            print(f"Playlist ID: {result['id']}\nPlaylist Title: {result['title']}")


@run.command(name="reset")
def reset_db():
    """Reset monitoring database"""
    logger.warning("** WARNING: All artists and playlists will be removed regardless of profile! **")
    confirm = input(":: Type 'reset' to confirm: ")
    if confirm.lower() == "reset":
        print("")
        db.reset_database()
    else:
        logger.info("Reset aborted. Database has NOT been modified.")
    return


@run.command(name='profile')
@click.argument('profile', required=False)
@click.option('-a', '--add', is_flag=True, help="Add new profile")
@click.option('-c', '--clear', is_flag=True, help="Clear config for existing profile")
@click.option('-d', '--delete', is_flag=True, help="Delete an existing profile")
@click.option('-e', '--edit', is_flag=True, help="Edit an existing profile")
def profile_command(profile, add, clear, delete, edit):
    """Add, modify and delete configuration profiles"""

    pc = ProfileConfig(profile)
    if profile:
        if add:
            pc.add()
        elif clear:
            pc.clear()
        elif delete:
            pc.delete()
        elif edit:
            pc.edit()
        else:
            pc.show()
    else:
        pc.show()


@run.command(name="search")
def search():
    """Interactively search and download/monitor artists"""
    client = Search()
    client.search_menu()


@run.command(name="config")
@click.argument('artist', nargs=-1, required=True)
def config_command(artist):
    """Configure per-artist settings by name or ID"""
    artist = ' '.join([x for x in artist])
    artist_lookup(artist)


@run.command(name="rollback", no_args_is_help=True)
@click.argument('num', type=int, required=False)
@click.option('-v', '--view', is_flag=True, help="View recent refresh transactions")
def rollback_command(num, view):
    """Rollback a previous monitor or refresh transaction"""
    if view:
        rollback.view_transactions()
    elif num:
        rollback.rollback_last(num)




import logging

from deemon.core.config import Config as config
from deemon.core.db import Database

logger = logging.getLogger(__name__)
db = Database()


def print_header(message: str = None):
    from os import system, name
    if name == 'nt':
        _ = system('cls')
    else:
        _ = system('clear')
    print("deemon Artist Configurator")
    if message:
        print(":: " + message + "\n")
    else:
        print("")


def get_artist(query: str):
    artist_as_id = False
    artist_fromdb = None

    try:
        artist_as_id = int(query)
    except ValueError:
        pass

    by_name = db.get_monitored_artist_by_name(query)
    if by_name:
        logger.debug("Artist found by name")
        artist_fromdb = by_name

    if artist_as_id:
        by_id = db.get_monitored_artist_by_id(artist_as_id)
        if by_id:
            logger.debug(f"Artist found by ID")
            if not artist_fromdb:
                artist_fromdb = by_id
            else:
                logger.debug("Artist found by both ID and name, prompting user")

                while True:
                    prompt = input(":: Multiple artists found. Was that a name or ID? ")
                    if prompt.lower() == "name":
                        logger.debug("Artist confirmed by user to be a name")
                        return artist_fromdb
                    elif prompt.lower() == "id":
                        logger.debug("Artist confirmed by user to be an ID")
                        return by_id
        else:
            return logger.error(f"Artist/Artist ID not found for '{query}'")

    if artist_fromdb:
        return artist_fromdb
    else:
        return logger.error(f"Artist/Artist ID not found for '{query}'")


def artist_lookup(query):
    result = get_artist(query)
    if not result:
        return
    print_header(f"Configuring '{result['artist_name']}' (Artist ID: {result['artist_id']})")
    modified = 0
    for property in result:
        if property not in ['alerts', 'bitrate', 'record_type', 'download_path']:
            continue
        allowed_opts = config.allowed_values(property)
        if isinstance(allowed_opts, dict):
            allowed_opts = [str(x.lower()) for x in allowed_opts.values()]

        while True:
            friendly_text = property.replace("_", " ").title()
            user_input = input(f"{friendly_text} [{result[property]}]: ").lower()
            if user_input == "":
                break
            elif user_input == "false" or user_input == "0":
                user_input = False
            elif user_input == "true" or user_input == "1":
                user_input = True
            if user_input == "none":
                user_input = None
            elif allowed_opts:
                if user_input not in allowed_opts:
                    print(f"Allowed options: " + ', '.join(str(x) for x in allowed_opts))
                    continue
            logger.debug(f"User set {property} to {user_input}")
            result[property] = user_input
            modified += 1
            break
    if modified > 0:
        i = input("\n:: Save these settings? [y|N] ")
        if i.lower() != "y":
            logger.info("No changes made, exiting...")
        else:
            db.update_artist(result)
            print(f"\nArtist '{result['artist_name']}' has been updated!")
    else:
        print("No changes made, exiting...")
